# coding: utf8

__all__ = ["BondRoundingParameters"]

from .._object_definition import ObjectDefinition
from .._enums import RoundingType, Rounding


class BondRoundingParameters(ObjectDefinition):
    def __init__(
        self,
        accrued_rounding=None,
        accrued_rounding_type=None,
        price_rounding=None,
        price_rounding_type=None,
        spread_rounding=None,
        spread_rounding_type=None,
        yield_rounding=None,
        yield_rounding_type=None,
    ):
        super().__init__()
        self.accrued_rounding = accrued_rounding
        self.accrued_rounding_type = accrued_rounding_type
        self.price_rounding = price_rounding
        self.price_rounding_type = price_rounding_type
        self.spread_rounding = spread_rounding
        self.spread_rounding_type = spread_rounding_type
        self.yield_rounding = yield_rounding
        self.yield_rounding_type = yield_rounding_type

    @property
    def accrued_rounding(self):
        """
        Number of digits to apply for rounding of Accrued field. Available values are Zero, One, Two,..., Eight, Default, Unrounded.
        Optional. A default value may be defined in bond reference data, in that case it is used. If it is not the case no rounding
        is applied.
        :return: enum Rounding
        """
        return self._get_enum_parameter(Rounding, "accruedRounding")

    @accrued_rounding.setter
    def accrued_rounding(self, value):
        self._set_enum_parameter(Rounding, "accruedRounding", value)

    @property
    def accrued_rounding_type(self):
        """
        Type of rounding for accrued rounding.
        Optional. A default value can be defined in bond reference data. Otherwise, default value is Near.
        :return: enum RoundingType
        """
        return self._get_enum_parameter(RoundingType, "accruedRoundingType")

    @accrued_rounding_type.setter
    def accrued_rounding_type(self, value):
        self._set_enum_parameter(RoundingType, "accruedRoundingType", value)

    @property
    def price_rounding(self):
        """
        Number of digits to apply for price rounding. Available values are Zero, One, Two,..., Eight, Default, Unrounded.
        Optional. A default value may be defined in bond reference data, in that case it is used. If it is not the case no rounding
        is applied.
        :return: enum Rounding
        """
        return self._get_enum_parameter(Rounding, "priceRounding")

    @price_rounding.setter
    def price_rounding(self, value):
        self._set_enum_parameter(Rounding, "priceRounding", value)

    @property
    def price_rounding_type(self):
        """
        Type of rounding for price rounding.
        Optional. A default value can be defined in bond reference data. Otherwise, default value is Near.
        :return: enum RoundingType
        """
        return self._get_enum_parameter(RoundingType, "priceRoundingType")

    @price_rounding_type.setter
    def price_rounding_type(self, value):
        self._set_enum_parameter(RoundingType, "priceRoundingType", value)

    @property
    def spread_rounding(self):
        """
        Number of digits to apply for spread rounding. Available values are Zero, One, Two,..., Eight, Default, Unrounded.
        Note that spread rounding is done directly on the base point value.
        Optional. By default, data from the bond structure.
        :return: enum Rounding
        """
        return self._get_enum_parameter(Rounding, "spreadRounding")

    @spread_rounding.setter
    def spread_rounding(self, value):
        self._set_enum_parameter(Rounding, "spreadRounding", value)

    @property
    def spread_rounding_type(self):
        """
        :return: enum RoundingType
        """
        return self._get_enum_parameter(RoundingType, "spreadRoundingType")

    @spread_rounding_type.setter
    def spread_rounding_type(self, value):
        self._set_enum_parameter(RoundingType, "spreadRoundingType", value)

    @property
    def yield_rounding(self):
        """
        Number of digits to apply for yield rounding. Available values are Zero, One, Two,..., Eight, Default, Unrounded.
        Optional. A default value may be defined in bond reference data, in that case it is used. If it is not the case no rounding
        is applied.
        :return: enum Rounding
        """
        return self._get_enum_parameter(Rounding, "yieldRounding")

    @yield_rounding.setter
    def yield_rounding(self, value):
        self._set_enum_parameter(Rounding, "yieldRounding", value)

    @property
    def yield_rounding_type(self):
        """
        Type of rounding for yield rounding.
        Optional. A default value can be defined in bond reference data. Otherwise, default value is Near.
        :return: enum RoundingType
        """
        return self._get_enum_parameter(RoundingType, "yieldRoundingType")

    @yield_rounding_type.setter
    def yield_rounding_type(self, value):
        self._set_enum_parameter(RoundingType, "yieldRoundingType", value)



# coding: utf-8

from ._search_view_type import SearchViews
from .._content_provider import ContentProviderLayer
from ..._tools import create_repr, validate_types


class Definition(ContentProviderLayer):
    """
    This class describe parameters to retrieve data for search.

    Parameters
    ----------
    query: str, optional
        Keyword argument for view

    view: SearchViews, optional
        The view for searching see at SearchViews enum.
        Default: SearchViews.SEARCH_ALL

    filter: str, optional
        Where query is for unstructured end-user-oriented restriction, filter is for
        structured programmatic restriction.

    order_by: str, optional
        Defines the order in which matching documents should be returned.

    boost: str, optional
        This argument supports exactly the same predicate expression syntax as filter,
        but where filter restricts which documents are matched at all,
        boost just applies a large scoring boost to documents it matches,
        which will almost always guarantee that they appear at the top of the results.

    select: str, optional
        A comma-separated list of the properties of a document to be returned in the response.

    top: int, optional
        the maximum number of documents to retrieve. Must be non-negative.
        default: 10

    skip: int, optional
        The number of documents to skip in the sorted result set before returning the
        next top.

    group_by: str, optional
        If specified, this must name a single Groupable property.
        returned documents are grouped into buckets based on their value for this
        property.

    group_count: str, optional
        When supplied in combination with group_by, sets the maximum number of documents
        to be returned per bucket.
        default: 3

    navigators: str, optional
        This can name one or more properties, separated by commas, each of which must
        be Navigable.

    extended_params : dict, optional
        other parameters can be provided if necessary

    Examples
    --------
    >>> from refinitiv.data.content import search
    >>> definition = search.Definition(query="cfo", view=search.SearchViews.PEOPLE)
    """

    def __init__(
        self,
        query: str = None,
        view: SearchViews = SearchViews.SEARCH_ALL,
        filter: str = None,
        order_by: str = None,
        boost: str = None,
        select: str = None,
        top: int = 10,
        skip: int = 0,
        group_by: str = None,
        group_count: int = 3,
        navigators: str = None,
        features: str = None,
        scope: str = None,
        terms: str = None,
        extended_params: dict = None,
    ):
        validate_types(group_count, [int], "group_count")
        validate_types(skip, [int], "skip")
        validate_types(top, [int], "top")

        self._query = query
        self._view = view
        self._boost = boost
        self._features = features
        self._filter = filter
        self._group_by = group_by
        self._group_count = group_count
        self._navigators = navigators
        self._order_by = order_by
        self._scope = scope
        self._select = select
        self._skip = skip
        self._terms = terms
        self._top = top
        self._extended_params = extended_params
        from .._content_type import ContentType

        super().__init__(
            content_type=ContentType.DISCOVERY_SEARCH,
            query=self._query,
            view=self._view,
            filter=self._filter,
            order_by=self._order_by,
            boost=self._boost,
            select=self._select,
            top=self._top,
            skip=self._skip,
            group_by=self._group_by,
            group_count=self._group_count,
            navigators=self._navigators,
            features=self._features,
            scope=self._scope,
            terms=self._terms,
            extended_params=self._extended_params,
        )

    def __repr__(self):
        return create_repr(self, content=f"{{query='{self._query}'}}")



# coding: utf8
# contract_gen 2020-05-18 08:30:59.243815

__all__ = ["NumeraireType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class NumeraireType(Enum):
    CASH = "Cash"
    ROLLING_EVENT = "RollingEvent"
    ROLLING_PAYMENT = "RollingPayment"
    TERMINAL_EVENT_ZC = "TerminalEventZc"
    TERMINAL_ZC = "TerminalZc"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(NumeraireType, _NUMERAIRE_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_NUMERAIRE_TYPE_VALUES_IN_LOWER_BY_NUMERAIRE_TYPE, some)


_NUMERAIRE_TYPE_VALUES = tuple(t.value for t in NumeraireType)
_NUMERAIRE_TYPE_VALUES_IN_LOWER_BY_NUMERAIRE_TYPE = {
    name.lower(): item for name, item in NumeraireType.__members__.items()
}



from typing import Iterable, Union, TYPE_CHECKING

from ..._content_provider import ContentProviderLayer
from ..._content_type import ContentType
from ...._tools import validate_types, validate_bool_value

if TYPE_CHECKING:
    from ....content._types import ExtendedParams


class Definition(ContentProviderLayer):
    """
    This class describe parameters to retrieve the calculated concentration data
    by top 10, 20, 50, 100 consolidated investors.

    Parameters
    ----------
    universe: str, list of str
        The Universe parameter allows the user to define the companies for which the content is returned.

    count: int
        Number of records

    use_field_names_in_headers: bool, optional
        Return field name as column headers for data instead of title

    extended_params : ExtendedParams, optional
        If necessary other parameters.

    Examples
    --------
    >>> from refinitiv.data.content import ownership
    >>> definition = ownership.consolidated.top_n_concentration.Definition("TRI.N", 30)
    >>> response = definition.get_data()
    """

    def __init__(
        self,
        universe: Union[str, Iterable[str]],
        count: int,
        use_field_names_in_headers: bool = False,
        extended_params: "ExtendedParams" = None,
    ):
        validate_types(count, [int], "count")
        validate_bool_value(use_field_names_in_headers)

        super().__init__(
            content_type=ContentType.OWNERSHIP_CONSOLIDATED_TOP_N_CONCENTRATION,
            universe=universe,
            count=count,
            use_field_names_in_headers=use_field_names_in_headers,
            extended_params=extended_params,
        )



import abc
import collections
import json
import os
import pathlib
from typing import List, Optional, TYPE_CHECKING, Union

from .proxy_info import ProxyInfo
from ..._configure import keys
from ..._core.session import SessionType
from ..._core.session._session_cxn_type import SessionCxnType
from ..._tools import parse_url, urljoin
from ...errors import RDError

if TYPE_CHECKING:
    from ..._core.session import Session, PlatformSession
    from ..._configure import _RDPConfig

StreamServiceInfo = collections.namedtuple(
    "StreamServiceInfo",
    ["scheme", "host", "port", "path", "data_formats", "location", "transport"],
)

_DEFAULT_RECONNECTION_DELAY_SECS = 5


class StreamCxnConfig(abc.ABC):
    def __init__(
        self,
        infos: Union[List["StreamServiceInfo"], "StreamServiceInfo"],
        protocols: Union[List[str], str],
        transport: str = "websocket",
    ):
        if isinstance(infos, list) and len(infos) == 0:
            raise ValueError("infos are empty")

        if not isinstance(infos, list):
            infos = [infos]

        if not isinstance(protocols, list):
            protocols = [protocols]

        self._infos = infos
        self._protocols = protocols
        self._transport = transport
        self._index = 0

    @property
    def transport(self):
        return self._transport

    @property
    def info(self):
        return self._infos[self._index]

    @property
    def url(self):
        return self._get_url(self.info)

    @property
    def url_scheme(self):
        return self.info.scheme

    @property
    def urls(self):
        return [self._get_url(info) for info in self._infos]

    @property
    def headers(self):
        return []

    @property
    def data_formats(self):
        return self.info.data_formats

    @property
    def supported_protocols(self):
        return self._protocols

    def reset_reconnection_config(self):
        self._index = 0

    @property
    def no_proxy(self):
        return ProxyInfo.get_no_proxy()

    @property
    def proxy_config(self):
        proxies_info = ProxyInfo.get_proxies_info()
        if self.url_scheme == "wss":
            # try to get https proxy then http proxy if https not configured
            return proxies_info.get("https", proxies_info.get("http", None))
        else:
            return proxies_info.get("http", None)

    @property
    def data_fmt(self):
        if not self.data_formats:
            return ""
        return self.data_formats[0]

    @property
    def delay(self):
        return self._index * _DEFAULT_RECONNECTION_DELAY_SECS

    def set_next_url(self):
        self._index = (self._index + 1) % len(self._infos)

    @abc.abstractmethod
    def _get_url(self, info: "StreamServiceInfo") -> str:
        pass

    def __str__(self) -> str:
        urls = "\n\t\t\t ".join(self.urls)
        s = (
            f"{self.__class__.__name__} {{\n"
            f"\t\tinfo={self.info},\n"
            f"\t\turl={self.url},\n"
            f"\t\turl_scheme={self.url_scheme},\n"
            f"\t\turls={urls},\n"
            f"\t\theaders={self.headers}, "
            f"data_formats={self.data_formats}, "
            f"supported_protocols={self.supported_protocols}, "
            f"no_proxy={self.no_proxy}, "
            f"proxy_config={self.proxy_config}, "
            f"data_fmt={self.data_fmt}, "
            f"delay={self.delay}}}"
        )
        return s


class DesktopStreamCxnConfig(StreamCxnConfig):
    def __init__(
        self,
        session: "Session",
        infos: Union[List["StreamServiceInfo"], "StreamServiceInfo"],
        protocols: Union[List[str], str],
    ):
        super().__init__(infos, protocols)
        self._session = session

    @property
    def headers(self):
        app_version = os.getenv("DP_PROXY_APP_VERSION")
        headers = [f"x-tr-applicationid: {self._session.app_key}"]

        if self._session._access_token:
            headers.append(f"Authorization: Bearer {self._session._access_token}")

        if app_version and self._session.type == SessionType.DESKTOP:
            headers.append(f"app-version: {app_version}")

        return headers

    def _get_url(self, info: "StreamServiceInfo") -> str:
        return f"{info.scheme}://{info.host}:{info.port}/{info.path}"


class PlatformStreamCxnConfig(StreamCxnConfig):
    def _get_url(self, info: "StreamServiceInfo") -> str:
        if self.transport == "tcp":
            return f"{info.host}:{info.port}"
        else:
            path = info.path or "WebSocket"
            return f"{info.scheme}://{info.host}:{info.port}/{path}"


class NullStreamCxnConfig(StreamCxnConfig):
    def __init__(self):
        StreamCxnConfig.__init__(
            self, [StreamServiceInfo("", "", "", "", "", "", "")], []
        )

    def _get_url(self, info):
        return ""


def get_discovery_url(
    root_url: str, streaming_name: str, endpoint_name: str, config: "_RDPConfig"
) -> str:
    config_name = f"apis.streaming.{streaming_name}"
    config_endpoint_name = f"{config_name}.endpoints.{endpoint_name}"
    base_path = config.get_str(f"{config_name}.url")

    try:
        endpoint_path = config.get_str(f"{config_endpoint_name}.path")
    except KeyError:
        raise KeyError(
            f"Cannot find discovery endpoint '{endpoint_name}' "
            f"for streaming '{streaming_name}' in config."
        )

    if base_path.startswith("http"):
        url = base_path
    else:
        url = urljoin(root_url, base_path)

    return urljoin(url, endpoint_path)


def _filter_by_location(locations: List[str], infos: List[StreamServiceInfo]) -> list:
    if not locations:
        return infos

    filtered = []
    for location in locations:
        for info in infos:
            has_location = any(
                loc.strip().startswith(location) for loc in info.location
            )
            if has_location and info not in filtered:
                filtered.append(info)

    return filtered


class CxnConfigProvider(abc.ABC):
    config_class = None
    _port_by_prefix = {
        80: "ws",
        443: "wss",
    }

    def get_cfg(
        self, session: "Session", api_cfg_key: str
    ) -> Union[PlatformStreamCxnConfig, DesktopStreamCxnConfig]:
        """
        Parameters
        ----------
        session: Session
        api_cfg_key: str
            Example - "streaming/pricing/main"

        Returns
        -------
        PlatformStreamCxnConfig or DesktopStreamCxnConfig

        """
        _, content_name, endpoint_name = api_cfg_key.split("/")
        cfg: "_RDPConfig" = session.config
        url: Optional[str] = None

        if cfg.get(f"apis.streaming.{content_name}.use_rwf", False):
            transport = "tcp"
        else:
            transport = "websocket"

        if transport == "websocket":
            url = cfg.get(keys.get_stream_websocket_url(content_name, endpoint_name))
        if transport == "tcp" or (transport == "websocket" and url is None):
            url = cfg.get(keys.get_stream_direct_url(content_name, endpoint_name))

        if url is not None:
            infos = [self.info_from_url(transport, url)]

        else:
            url_root: str = session._get_rdp_url_root()
            discovery_url: str = get_discovery_url(
                url_root, content_name, endpoint_name, cfg
            )
            infos = self._request_infos(
                discovery_url, api_cfg_key, cfg, session, transport
            )
        protocols = cfg.get_list(keys.stream_protocols(content_name, endpoint_name))
        return self._create_cfg(session, infos, protocols, transport)

    @staticmethod
    def info_from_url(transport: str, url: str, data_formats=None) -> StreamServiceInfo:
        if data_formats is None:
            data_formats = ["unknown"]

        # If there is no scheme or netloc add netloc marker to make it valid URL
        if not ("://" in url or url.startswith("//")) and not url.startswith("/"):
            url = "//" + url

        result = parse_url(url)
        scheme = result.scheme
        host = result.hostname
        port = result.port
        path = result.path

        # If url parsing did not get valid hostname, raise exception
        if not host:
            raise ValueError(f"Invalid URL: {url}")

        if not scheme and transport == "websocket":
            scheme = "wss" if port == 443 else "ws"
        return StreamServiceInfo(
            scheme=scheme or "",
            host=host or "",
            port=port or 80,
            path=path or "",
            data_formats=data_formats,
            location="",
            transport=transport,
        )

    def _request_infos(
        self,
        discovery_url: str,
        api_config_key: str,
        config: "_RDPConfig",
        session: "Session",
        transport: str = "websocket",
    ) -> List[StreamServiceInfo]:

        content_name = api_config_key.split("/")[1]
        tier: Optional[int] = config.get(f"apis.streaming.{content_name}.tier")

        response = session.http_request(
            discovery_url,
            # server won't accept tier: false
            params={"tier": True} if tier else {},
        )
        try:
            data = response.json()
        except json.decoder.JSONDecodeError:
            raise RuntimeError(f"Cannot load config from {discovery_url}, {response}")

        err = data.get("error")
        if err:
            raise RDError(response.status_code, err.get("message"))

        infos = []
        for service in data.get("services", []):
            if service.get("transport") != transport:
                continue

            if tier is not None and "tier" in service:
                tier_range: List[int] = service["tier"]
                if tier < tier_range[0] or tier > tier_range[1]:
                    continue

            endpoint_path = pathlib.Path(service.get("endpoint"))
            host = str(endpoint_path.parts[0])
            path = "/".join(endpoint_path.parts[1:])

            scheme = ""
            port = service.get("port")
            if transport == "websocket":
                scheme = self._port_by_prefix.get(port, "ws")

            infos.append(
                StreamServiceInfo(
                    scheme=scheme,
                    host=host,
                    port=port,
                    path=path,
                    data_formats=service.get("dataFormat", ["unknown"]),
                    location=service.get("location"),
                    transport=transport,
                )
            )

        return self._filter_infos(infos, api_config_key, config)

    def _filter_infos(
        self,
        infos: List[StreamServiceInfo],
        api_cfg_key: str,
        cfg: "_RDPConfig",
    ) -> List[StreamServiceInfo]:
        return infos

    def _create_cfg(
        self,
        session: "Session",
        infos: List[StreamServiceInfo],
        protocols: List[str],
        transport: str = "websocket",
    ) -> Union[PlatformStreamCxnConfig, DesktopStreamCxnConfig]:
        return self.config_class(infos, protocols, transport)


class DesktopCxnConfigProvider(CxnConfigProvider):
    config_class = DesktopStreamCxnConfig

    def _create_cfg(
        self,
        session: "Session",
        infos: List[StreamServiceInfo],
        protocols: List[str],
        transport: str = "websocket",
    ) -> Union[PlatformStreamCxnConfig, DesktopStreamCxnConfig]:
        return self.config_class(session, infos, protocols)


class PlatformCxnConfigProvider(CxnConfigProvider):
    config_class = PlatformStreamCxnConfig

    def _filter_infos(
        self,
        infos: List[StreamServiceInfo],
        api_cfg_key: str,
        cfg: "_RDPConfig",
    ) -> List[StreamServiceInfo]:
        _, content_name, endpoint_name = api_cfg_key.split("/")

        locations = cfg.get_list(
            keys.stream_connects_locations(content_name, endpoint_name)
        )
        return _filter_by_location(locations, infos)


class DeployedCxnConfigProvider(CxnConfigProvider):
    def get_cfg(
        self, session: "PlatformSession", api_cfg_key: str
    ) -> PlatformStreamCxnConfig:

        url: str = session._deployed_platform_host
        cfg: "_RDPConfig" = session.config
        _, content_name, endpoint_name = api_cfg_key.split("/")

        if url is None:
            session_name: str = session.name
            key = keys.platform_realtime_distribution_system(session_name)
            url_key = f"{key}.url"
            url = cfg.get_str(url_key)

        if cfg.get(f"apis.streaming.{content_name}.use_rwf", False):
            transport = "tcp"
            data_formats = None
        else:
            transport = "websocket"
            data_formats = ["tr_json2"]

        info = self.info_from_url(transport, url, data_formats=data_formats)

        return PlatformStreamCxnConfig(info, "OMM", transport=transport)


class PlatformAndDeployedCxnConfigProvider(
    DeployedCxnConfigProvider, PlatformCxnConfigProvider
):
    def get_cfg(self, session: "PlatformSession", api_cfg_key: str) -> StreamCxnConfig:

        if api_cfg_key.startswith("streaming/pricing/main"):
            cxn_config = DeployedCxnConfigProvider.get_cfg(self, session, api_cfg_key)
        else:
            cxn_config = PlatformCxnConfigProvider.get_cfg(self, session, api_cfg_key)

        return cxn_config


cxn_cfg_provider_by_session_cxn_type = {
    SessionCxnType.DEPLOYED: DeployedCxnConfigProvider(),
    SessionCxnType.REFINITIV_DATA: PlatformCxnConfigProvider(),
    SessionCxnType.REFINITIV_DATA_AND_DEPLOYED: PlatformAndDeployedCxnConfigProvider(),
    SessionCxnType.DESKTOP: DesktopCxnConfigProvider(),
}


def get_cxn_config(api_config_key: str, session: "Session") -> StreamCxnConfig:
    session_cxn_type = session._get_session_cxn_type()
    cxn_cfg_provider = cxn_cfg_provider_by_session_cxn_type.get(session_cxn_type)

    if not cxn_cfg_provider:
        raise ValueError(
            f"Can't find cxn_cfg_provider by session_cxn_type={session_cxn_type}"
        )

    return cxn_cfg_provider.get_cfg(session, api_config_key)



# coding: utf-8

import json
import time

from httpx import ConnectError
from .._tools._common import get_response_reason

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError

import sys
from ._tools import is_string_type, get_default_session
from ..errors import RDError


def send_json_request(entity, payload, ID="123", debug=False, raw_response=False):
    """
    Returns the JSON response.
    This legacy can be used for advanced usage or early access to new features.

    Parameters
    ----------
    entity: string
        A string containing a service name

    payload: string, dict
        A string containing a JSON request

    debug: boolean, optional
        When set to True, the json request and response are printed.
        Default: False

    Returns
    -------
    string
        The JSON response as a string

    Raises
    ------
    ElektronError

        If daemon is disconnected

    requests.Timeout
        If request times out

    Exception
        If request fails (HTTP code other than 200)

    ElektronError
        If daemon is disconnected
    """
    _session = get_default_session()
    if _session:
        logger = _session.logger()
        logger.debug("entity: {}".format(entity))
        logger.debug("payload: {}".format(payload))

        if not is_string_type(entity):
            error_msg = "entity must be a string identifying an UDF endpoint"
            logger.error(error_msg)
            raise ValueError(error_msg)
        try:
            if is_string_type(payload):
                data = json.loads(payload)
            elif type(payload) is dict:
                data = payload
            else:
                error_msg = "payload must be a string or a dictionary"
                logger.error(error_msg)
                raise ValueError(error_msg)
        except JSONDecodeError as e:
            error_msg = "payload must be json well formed.\n"
            error_msg += str(e)
            logger.error(error_msg)
            raise e

        try:
            # build the request
            udf_request = {"Entity": {"E": entity, "W": data}, "ID": ID}
            logger.debug(
                "Request to {} :{}".format(_session._get_udf_url(), udf_request)
            )
            response = _session.http_request(
                method="POST",
                url=_session._get_udf_url(),
                headers={"Content-Type": "application/json"},
                # "x-tr-applicationid": _session.app_key},
                json=udf_request,
            )

            try:
                logger.debug("HTTP Response code: {}".format(response.status_code))
                logger.debug("HTTP Response: {}".format(response.text))
            except UnicodeEncodeError:
                logger.error("HTTP Response: cannot decode error message")

            if response.status_code == 200:
                result = {}
                try:
                    result = response.json()
                    logger.debug(
                        "Response size: {}".format(sys.getsizeof(json.dumps(result)))
                    )
                except JSONDecodeError:
                    logger.error(f"Failed to decode response to json: {response.text}")

                # Manage specifically DataGrid async mode
                if entity.startswith("DataGrid") and entity.endswith("Async"):
                    ticket = _check_ticket_async(result)
                    while ticket:
                        ticket_request = {
                            "Entity": {
                                "E": entity,
                                "W": {"requests": [{"ticket": ticket}]},
                            }
                        }
                        logger.debug("Send ticket request:{}".format(ticket_request))
                        response = _session.http_request(
                            method="POST",
                            url=_session._get_udf_url(),
                            headers={"Content-Type": "application/json"},
                            json=ticket_request,
                        )

                        result = response.json()
                        logger.debug(
                            "Response size: {}".format(
                                sys.getsizeof(json.dumps(result))
                            )
                        )
                        ticket = _check_ticket_async(result)

                if not raw_response:
                    check_server_error(result)
                return response
            else:
                raise_for_status(response)

        except ConnectError as connectionError:
            error_msg = "Eikon Proxy not installed or not running. Please read the documentation to know how to install and run the proxy"
            logger.error(error_msg)
            raise RDError(401, error_msg) from connectionError


def _check_ticket_async(server_response):
    """
    Check server response.

    Check is the server response contains a ticket.

    :param server_response: request's response
    :type server_response: requests.Response
    :return: ticket value if response contains a ticket, None otherwise
    """
    logger = get_default_session().logger()
    # ticket response should contains only one key
    if len(server_response) == 1:
        for key, value in server_response.items():
            ticket = value[0]
            if ticket and ticket.get("estimatedDuration"):
                ticket_duration = int(ticket["estimatedDuration"])
                ticket_duration = min(ticket_duration, 15000)
                ticket_value = ticket["ticket"]
                message = "Receive ticket from {}, wait for {} second".format(
                    key, ticket_duration / 1000.0
                )
                if ticket_duration > 1000:
                    message = message + "s"
                logger.info(message)
                time.sleep(ticket_duration / 1000.0)
                return ticket_value
    return None


def check_server_error(server_response, session=None):
    """
    Check server response.

    Check is the server response contains an HTPP error or a server error.

    :param server_response: request's response
    :type server_response: requests.Response
    :return: nothing

    :raises: Exception('HTTP error : <_message>) if response contains HTTP response
              ex: '<500 Server error>'
          or Exception('Server error (<error code>) : <server_response>') if UDF returns an error
              ex: {u'ErrorCode': 500, u'ErrorMessage': u'Requested datapoint was not found: News_Headlines', u'Id': u''}

    """
    if session:
        logger = session.logger()
    else:
        logger = get_default_session().logger()

    # check HTTP response (server response is an object that can contain ErrorCode attribute)
    if hasattr(server_response, "ErrorCode"):
        logger.error(getattr(server_response, "ErrorMessage"))
        raise RDError(
            server_response["ErrorCode"], getattr(server_response, "ErrorMessage")
        )

    # check UDF response (server response is JSON and it can contain ErrorCode + ErrorMessage keys)
    if (
        isinstance(server_response, dict)
        and "ErrorCode" in server_response
        and "ErrorMessage" in server_response
    ):
        error_message = server_response["ErrorMessage"]
        logger.error(error_message)
        raise RDError(server_response["ErrorCode"], error_message)

    # check DataGrid response (server response is JSON or text and it can contain error + optionally transactionId keys)
    if "error" in server_response:
        if "transactionId" in server_response:
            error_message = f"{server_response['error']} (transactionId:{server_response['transactionId']}"
        else:
            error_message = f"{server_response['error']} (no transactionId)"
        logger.error(error_message)
        raise RDError(400, error_message)


def raise_for_status(response):
    """Raises stored :class:`HTTPError`, if one occurred."""

    error_msg = ""
    if isinstance(response, bytes):
        # We attempt to decode utf-8 first because some servers
        # choose to localize their reason strings. If the string
        # isn't utf-8, we fall back to iso-8859-1 for all other
        # encodings. (See PR #3538)
        try:
            reason = get_response_reason(response).decode("utf-8")
        except UnicodeDecodeError:
            reason = get_response_reason(response).decode("iso-8859-1")
    else:
        reason = get_response_reason(response)

    logger = get_default_session().logger()

    if 400 <= response.status_code < 500:
        error_msg = "Client Error: %s - %s" % (reason, response.text)
    elif 500 <= response.status_code < 600:
        error_msg = "Server Error: %s - %s" % (reason, response.text)

    if error_msg:
        logger.error("Error code {} | {}".format(response.status_code, error_msg))
        raise RDError(response.status_code, error_msg)



# coding: utf-8

__version__ = "1.0.59"

from ._forward_curve_definition import *
from ._forward_curve_request_item import *
from ._swap_zc_curve_definition import *
from ._swap_zc_curve_parameters import *
from ._interest_rate_curve_get_definition import *
from ._zc_curve_request_item import *
from ._zc_curve_definitions import *
from ._zc_curve_parameters import *

from ._enums import ZcCurvesOutputs



from typing import Optional, Callable, List

OptStr = Optional[str]
Strings = List[str]
Dicts = List[dict]
OptStrings = Optional[Strings]
OptDicts = Optional[Dicts]

OptInt = Optional[int]
OptFloat = Optional[float]
OptList = Optional[list]
OptTuple = Optional[tuple]
OptDict = Optional[dict]
OptSet = Optional[set]
OptBool = Optional[bool]
OptCall = Optional[Callable]

ExtendedParams = OptDict



import traceback
import urllib.parse
from json import JSONDecodeError
from typing import Any, Callable, Dict, Iterable, TYPE_CHECKING, Generic, TypeVar

import requests

from ._endpoint_data import EndpointData, Error, RequestMethod
from ..._configure import _RDPConfig
from ..._core.session import get_valid_session, is_open
from ..._tools import DEBUG, cached_property, get_response_reason, parse_url, urljoin
from ...content._content_type import ContentType
from ...errors import RDError

if TYPE_CHECKING:
    import pandas as pd


# --------------------------------------------------------------------------------------
#   Content data
# --------------------------------------------------------------------------------------


class Data(EndpointData):
    def __init__(
        self,
        raw: Any,
        dataframe: "pd.DataFrame" = None,
        dfbuilder: Callable[[Any, Dict[str, Any]], "pd.DataFrame"] = None,
        **kwargs,
    ):
        EndpointData.__init__(self, raw, **kwargs)
        self._dataframe = dataframe
        self._dfbuilder = dfbuilder

    @property
    def df(self):
        if self._dataframe is None and self._dfbuilder:
            self._dataframe = self._dfbuilder(self.raw, **self._kwargs)

        return self._dataframe


# --------------------------------------------------------------------------------------
#   Response object
# --------------------------------------------------------------------------------------

T = TypeVar("T")


class BaseResponse(Generic[T]):
    def __init__(self, *args, **kwargs):
        self.is_success: bool = kwargs.get("is_success")
        self.data: T = kwargs.get("content_data")
        self._status = kwargs.get("status")

        err_code = kwargs.get("error_code")
        err_message = kwargs.get("error_message")

        if err_code or err_message:
            if isinstance(err_code, list) and isinstance(err_message, list):
                errors = [Error(code, msg) for code, msg in zip(err_code, err_message)]

            else:
                errors = [Error(err_code, err_message)]

            self.errors = errors

        else:
            self.errors = []

        self._raw_response = kwargs.get("raw_response")
        self.http_response = self._raw_response

    @cached_property
    def requests_count(self):
        if isinstance(self.http_response, list):
            return len(self.http_response)
        return 1

    @property
    def request_message(self):
        if self._raw_response:
            return self._raw_response.request
        return None

    @property
    def closure(self):
        if self._raw_response:
            request = self._raw_response.request
            if isinstance(request, list):
                closure = [_request.headers.get("closure") for _request in request]
            else:
                closure = request.headers.get("closure")
            return closure
        return None

    @property
    def http_status(self):
        return self._status

    @property
    def http_headers(self):
        if self._raw_response:
            return self._raw_response.headers
        return None


class Response(BaseResponse[Data]):
    pass


# --------------------------------------------------------------------------------------
#   Response factory
# --------------------------------------------------------------------------------------


class ResponseFactory:
    def __init__(self, response_class=Response, data_class=Data):
        super().__init__()
        self.data_class = data_class
        self.response_class = response_class

    def get_raw(self, data):
        raw = data.get("content_data")
        return raw

    def get_dfbuilder(self, kwargs):
        from ...content._df_builder_factory import get_dfbuilder, DFBuildType

        content_type = kwargs.get("__content_type__", ContentType.DEFAULT)
        dfbuild_type = kwargs.get("__dfbuild_type__", DFBuildType.DATE_AS_INDEX)
        dfbuilder = get_dfbuilder(content_type, dfbuild_type)
        return dfbuilder

    def create_success(self, data, *args, **kwargs):
        inst = self.response_class(is_success=True, **data)
        raw = self.get_raw(data)
        dfbuilder = self.get_dfbuilder(kwargs)
        inst.data = self.data_class(raw, dfbuilder=dfbuilder, **kwargs)
        inst.data._owner = inst

        return inst

    def create_fail(self, data, *args, **kwargs):
        inst = self.response_class(is_success=False, **data)
        inst.data = self.data_class(data.get("content_data"))
        inst.data._owner = inst
        return inst


# --------------------------------------------------------------------------------------
#   Request factory
# --------------------------------------------------------------------------------------

request_method_value = {
    RequestMethod.POST: "POST",
    RequestMethod.PUT: "PUT",
    RequestMethod.DELETE: "DELETE",
    RequestMethod.GET: "GET",
}


class RequestFactory:
    def get_url(self, *args, **kwargs):
        url = args[1]
        return url

    def create(self, *args, **kwargs):
        session = args[0]
        url = self.get_url(*args, **kwargs)
        url_root = session._get_rdp_url_root()

        method = self.get_request_method(*args, **kwargs)
        header_parameters = self.get_header_parameters(*args, **kwargs) or {}
        path_parameters = self.get_path_parameters(*args, **kwargs)
        extended_params = kwargs.get("extended_params") or {}
        query_parameters = self.get_query_parameters(*args, **kwargs)
        query_parameters = self.extend_query_parameters(
            query_parameters, extended_params
        )
        body_parameters = self.get_body_parameters(*args, **kwargs)
        body_parameters = self.extend_body_parameters(body_parameters, extended_params)
        closure = kwargs.get("closure")

        url = self.update_url(url_root, url, path_parameters, query_parameters)

        headers = header_parameters
        if method != RequestMethod.GET:
            headers["Content-Type"] = "application/json"

        request = {
            "url": url,
            "method": request_method_value.get(method),
            "headers": headers,
            "body": body_parameters,
        }

        if closure is not None:
            request["closure"] = closure

        return request

    def update_url(self, url_root, url, path_parameters, query_parameters):
        try:
            result = parse_url(url)

            if not all([result.scheme, result.netloc, result.path]):
                if result.query and result.path:
                    url = urljoin(url_root, result.path)
                else:
                    url = urljoin(url_root, url)
            else:
                url = "".join([url_root, result.path])

            if result.query and isinstance(query_parameters, list):
                query_parameters.extend(urllib.parse.parse_qsl(result.query))

        except Exception:
            url = "/".join([url_root + "/" + url])

        if path_parameters:
            for key, value in path_parameters.items():
                value = urllib.parse.quote(value)
                url = url.replace("{" + key + "}", value)

        if self.is_multiple_query_parameters(query_parameters):
            url = self.add_multiple_query_parameters(url, query_parameters)
        elif query_parameters:
            url = self.add_query_parameters(url, query_parameters)

        return url

    def is_multiple_query_parameters(self, query_parameters: Any) -> bool:
        """Check multiple query parameters.

        Args:
            query_parameters (Any): Query parameters.

        Uses to check query parameters if they passes like
        {"universe": ["IBM.N", "ORCL.N"]} in case when query string must be constructed
        like '?universe=BNPP.PA,ORCL.N'
        """
        if isinstance(query_parameters, dict):
            return all(
                [
                    len(query_parameters) == 1,
                    isinstance(list(query_parameters.values())[0], list),
                ]
            )

        return False

    def get_request_method(self, *_, **kwargs) -> RequestMethod:
        return kwargs.get("method", RequestMethod.GET)

    def get_body_parameters(self, *_, **kwargs) -> dict:
        return kwargs.get("body_parameters") or {}

    def get_query_parameters(self, *_, **kwargs) -> list:
        return kwargs.get("query_parameters") or []

    def get_path_parameters(self, *_, **kwargs) -> dict:
        return kwargs.get("path_parameters") or {}

    def add_multiple_query_parameters(self, url, query_parameters) -> str:
        """Add multiple query parameters to query string.

        Args:
            url (str): url to construct query string.
            query_parameters (dict): query parameters to construct query string.

        Uses to construct percent-encoded query string if query parameters passed like
        {"universe": ["IBM.N", "ORCL.N"]} and must be transformed to query string like
        '?universe=BNPP.PA,ORCL.N', not like '?universe=BNPP.PA&universe=ORCL.N'

        Estimates API accepts only '?universe=BNPP.PA,ORCL.N' format to get estimates
        actual KPI measures for multiple universe.
        """
        key, v = query_parameters.popitem()
        val = ", ".join(v)
        return "?".join([url, urllib.parse.urlencode({key: val})])

    def add_query_parameters(self, url, query_parameters) -> str:
        return "?".join([url, urllib.parse.urlencode(query_parameters)])

    def extend_query_parameters(self, query_parameters, extended_params):
        return query_parameters

    def extend_body_parameters(self, body_parameters, extended_params):
        result = body_parameters
        result.update(extended_params or {})
        return result

    def get_header_parameters(self, *args, **kwargs):
        return kwargs.get("header_parameters", {})


# --------------------------------------------------------------------------------------
#   Connection object
# --------------------------------------------------------------------------------------


class HttpSessionConnection:
    def send(self, request, *args, **kwargs):
        session = args[0]

        if not is_open(session):
            raise ValueError("Session is not opened. Can't send any request")

        method = request.get("method")
        url = request.get("url")
        headers = request.get("headers")
        body = request.get("body")
        auto_retry = kwargs.get("auto_retry", False)
        session.debug(f"Send {method} request to {url}  (auto-retry={auto_retry})")
        session.debug(f"\theaders : {headers}\n\tbody : {body}")
        data = session.http_request(
            url=url,
            method=method,
            headers=headers,
            json=body,
            closure=request.get("closure"),
            auto_retry=auto_retry,
        )
        return data

    async def send_async(self, request, *args, **kwargs):
        session = args[0]

        if not is_open(session):
            raise ValueError("Session is not opened. Can't send any request")

        method = request.get("method")
        url = request.get("url")
        headers = request.get("headers")
        body = request.get("body")
        auto_retry = kwargs.get("auto_retry", False)
        session.debug(
            f"Send {method} request asynchronously to {url}  (auto-retry={auto_retry})"
        )
        session.debug(f"\theaders : {headers}\n\tbody : {body}")
        data = await session.http_request_async(
            url=url,
            method=method,
            headers=headers,
            json=body,
            closure=request.get("closure"),
            auto_retry=auto_retry,
        )
        return data


# --------------------------------------------------------------------------------------
#   Raw data parser
# --------------------------------------------------------------------------------------

success_http_codes = [
    requests.codes.ok,
    requests.codes.accepted,
    requests.codes.created,
]


class Parser:
    def process_successful_response(self, raw_response):
        status = {
            "http_status_code": raw_response.status_code,
            "http_reason": get_response_reason(raw_response),
        }

        media_type = raw_response.headers.get("content-type", "")

        if "/json" in media_type:
            try:
                content_data = raw_response.json()
                if content_data is None:
                    # Some HTTP responses, such as a DELETE,
                    # can be successful without any response body.
                    content_data = {}

                parsed_data = {
                    "status": status,
                    "raw_response": raw_response,
                    "content_data": content_data,
                }
            except (TypeError, JSONDecodeError) as error:
                message = f"Failed to process HTTP response : {str(error)}"
                status["content"] = message
                content_data = raw_response.text
                parsed_data = {
                    "status": status,
                    "raw_response": raw_response,
                    "content_data": content_data,
                }

        elif (
            "text/plain" in media_type
            or "text/html" in media_type
            or "text/xml" in media_type
            or "image/" in media_type
        ):
            content_data = raw_response.text
            parsed_data = {
                "status": status,
                "raw_response": raw_response,
                "content_data": content_data,
            }

        else:
            status["content"] = f"Unknown media type returned: {media_type}"
            parsed_data = {
                "status": status,
                "raw_response": raw_response,
            }

        return parsed_data

    def process_failed_response(self, raw_response):
        status = {
            "http_status_code": raw_response.status_code,
            "http_reason": get_response_reason(raw_response),
        }

        try:
            content_data = raw_response.json()
            content_error = content_data.get("error")

            if content_error:
                if isinstance(content_error, str):
                    content_error = {"code": None, "message": content_error}

                status["error"] = content_error
                error_code = raw_response.status_code
                error_message = content_error.get("message")
            else:
                error_code = raw_response.status_code
                error_message = raw_response.text

        except (TypeError, JSONDecodeError):
            error_code = raw_response.status_code
            error_message = raw_response.text

        if error_code == 403:
            if not error_message.endswith("."):
                error_message += ". "
            error_message += "Contact Refinitiv to check your permissions."

        parsed_data = {
            "status": status,
            "error_code": error_code,
            "error_message": error_message,
            "raw_response": raw_response,
        }

        return parsed_data

    def parse_raw_response(self, raw_response, *args, **kwargs):
        parsed_data = {}
        is_success = False

        if raw_response is None:
            return is_success, parsed_data

        is_success = raw_response.status_code in success_http_codes

        if is_success:
            parsed_data = self.process_successful_response(raw_response)

        else:
            parsed_data = self.process_failed_response(raw_response)

        return is_success, parsed_data


# --------------------------------------------------------------------------------------
#   Content data validator
# --------------------------------------------------------------------------------------


class ContentValidator:
    def validate(self, data: dict, *args, **kwargs) -> bool:
        is_valid = True
        content_data = data.get("content_data")

        if content_data is None:
            is_valid = False
            data["error_code"] = 1
            data["error_message"] = "Content data is None"

        else:
            status = content_data.get("status")
            if status == "Error":
                is_valid = False
                data["error_code"] = content_data.get("code", -1)
                data["error_message"] = content_data.get("message")

        return is_valid


class ContentTypeValidator:
    def __init__(self, allowed_content_types=None):
        if allowed_content_types is None:
            allowed_content_types = {"application/json"}
        self._allowed_content_types = allowed_content_types

    def validate(self, data: dict, *args, **kwargs) -> bool:
        # Checking only first part (type/subtype) of media_type
        # See https://httpwg.org/specs/rfc7231.html#media.type
        content_type = (
            data["raw_response"].headers.get("content-type", "").split(";")[0].strip()
        )
        is_success = content_type in self._allowed_content_types

        if not is_success:
            data["error_code"] = -1
            data["error_message"] = (
                f"Unexpected content-type in response,\n"
                f"Expected: {self._allowed_content_types}\n"
                f"Actual: {content_type}"
            )

        return is_success


# --------------------------------------------------------------------------------------
#   Data provider
# --------------------------------------------------------------------------------------


def _check_response(
    response: "Response", config: "_RDPConfig", response_class=Response
) -> None:
    if isinstance(response, response_class):
        is_raise_exception = config.get_param("raise_exception_on_error")
        if not response.is_success and is_raise_exception:
            error_code = response.errors[0].code
            error_message = response.errors[0].message
            exception_class = getattr(response, "exception_class", None)

            if exception_class:
                raise exception_class(error_code, error_message)

            else:
                raise RDError(error_code, error_message)


class ValidatorContainer:
    def __init__(
        self,
        validators: Iterable = None,
        content_validator=ContentValidator(),
        content_type_validator=ContentTypeValidator(),
        use_default_validators=True,
    ):
        self.validators = list(validators) if validators else []
        if content_validator and use_default_validators:
            self.validators.append(content_validator)
        if content_type_validator and use_default_validators:
            self.validators.append(content_type_validator)

    def validate(self, data: dict, *args, **kwargs) -> bool:
        return all(
            validator.validate(data, *args, **kwargs) for validator in self.validators
        )


class DataProvider:
    def __init__(
        self,
        connection=HttpSessionConnection(),
        request=RequestFactory(),
        response=ResponseFactory(),
        parser=Parser(),
        validator=ValidatorContainer(),
    ):
        self.connection = connection
        self.request = request
        self.response = response
        self.parser = parser
        self.validator = validator

    def _process_response(self, raw_response, *args, **kwargs):
        is_success, data = self.parser.parse_raw_response(raw_response)

        is_success = is_success and self.validator.validate(data, *args, **kwargs)

        if is_success:
            response = self.response.create_success(data, *args, **kwargs)

        else:
            response = self.response.create_fail(data, *args, **kwargs)

        return response

    def get_data(self, *args, **kwargs):
        request = self.request.create(*args, **kwargs)
        raw_response = self.connection.send(request, *args, **kwargs)
        return self._process_response(raw_response, *args, **kwargs)

    async def get_data_async(self, *args, **kwargs):
        request = self.request.create(*args, **kwargs)
        raw_response = await self.connection.send_async(request, *args, **kwargs)
        return self._process_response(raw_response, *args, **kwargs)


default_data_provider = DataProvider()


# --------------------------------------------------------------------------------------
#   Provider layer
# --------------------------------------------------------------------------------------


def emit_event(handler, *args, **kwargs):
    session = args[2]
    try:
        handler(*args, **kwargs)
    except Exception as e:
        session.error(f"{handler} callback raised exception: {e!r}")
        session.debug(traceback.format_exc())

        if DEBUG:
            raise e


class DataProviderLayer:
    def __init__(self, data_type, **kwargs):
        self._initialize(data_type, **kwargs)

    def _initialize(self, data_type, **kwargs):
        from ._data_provider_factory import make_provider

        self._kwargs = kwargs
        self._kwargs["__data_type__"] = data_type
        self._kwargs["__content_type__"] = data_type

        self._data_type = data_type
        self._content_type = data_type
        self._provider = make_provider(data_type)

    def _check_response(self, response, config):
        _check_response(response, config)

    def get_data(self, session=None, on_response=None, **kwargs):
        """
        Returns a response to the data platform

        Parameters
        ----------
        session : Session, optional
            Means default session would be used
        on_response : Callable, optional
            Callable object to process retrieved data

        Returns
        -------
        Response

        Raises
        ------
        AttributeError
            If user didn't set default session.

        """
        from ._data_provider_factory import get_url, get_api_config

        session = get_valid_session(session)
        config = session.config
        data_type = self._data_type
        url = get_url(data_type, config)
        api_config = get_api_config(data_type, config)
        auto_retry = api_config.get("auto-retry", False)
        response = self._provider.get_data(
            session, url, auto_retry=auto_retry, **kwargs, **self._kwargs
        )
        on_response and emit_event(on_response, response, self, session)
        self._check_response(response, config)
        return response

    async def get_data_async(self, session=None, on_response=None, **kwargs):
        """
        Returns a response asynchronously to the data platform

        Parameters
        ----------
        session : Session, optional
            Means default session would be used
        on_response : Callable, optional
            Callable object to process retrieved data

        Returns
        -------
        Response

        Raises
        ------
        AttributeError
            If user didn't set default session.

        """
        from ._data_provider_factory import get_url, get_api_config

        session = get_valid_session(session)
        config = session.config
        data_type = self._data_type
        url = get_url(data_type, config)
        api_config = get_api_config(data_type, config)
        auto_retry = api_config.get("auto-retry", False)
        response = await self._provider.get_data_async(
            session, url, auto_retry=auto_retry, **kwargs, **self._kwargs
        )
        on_response and emit_event(on_response, response, self, session)
        return response

    def __repr__(self):
        s = super().__repr__()
        s = s.replace(">", f" {{name='{self._kwargs.get('universe')}'}}>")
        return s



from typing import Callable

from . import StreamConnection
from ._stream_cxn_config import NullStreamCxnConfig
from ..._core.session.null_session import NullSession


class NullStreamConnection(StreamConnection):
    def __init__(self):
        StreamConnection.__init__(
            self,
            connection_id=-1,
            name="",
            session=NullSession(),
            config=NullStreamCxnConfig(),
            max_reconnect=0,
        )

    @property
    def can_run(self):
        return False

    @property
    def is_ready(self):
        return False

    @property
    def is_disposed(self):
        return False

    @property
    def can_be_disconnect(self):
        return False

    def connect(self):
        # do nothing
        pass

    def disconnect(self):
        # do nothing
        pass

    def dispose(self):
        # do nothing
        pass

    def send(self, data: dict) -> None:
        # do nothing
        pass

    def on(self, event: str, listener: Callable) -> None:
        # do nothing
        pass

    def remove_listener(self, event: str, listener: Callable) -> None:
        # do nothing
        pass



# coding: utf-8

__all__ = ("Definition",)


class Definition(object):
    """
    Create definition object for desktop session.
    When user provides "session-name", then it is checked
    in the config file for matches, if match is found,
    then this session will be taken, otherwise, an exception will be raised

    When user doesn't provide a session name,
    then the default session from the config file is taken

    Parameters
    ----------
        name: str
            name of the session in config. For example: 'default'
        app_key: str
            application key
    Raises
    ---------
    Exception
        If app-key is not found in the config file and in arguments.

    Examples
    --------
    >>> from refinitiv.data import session
    >>> definition = session.desktop.Definition(name="custom-session-name")
    >>> desktop_session = definition.get_session()
    """

    def __init__(
        self,
        name: str = "workspace",
        app_key: str = None,
    ):
        from .._core.session._session_provider import (
            _make_desktop_session_provider_by_arguments,
        )

        if not isinstance(name, str):
            raise ValueError("Invalid session name type, please provide string.")

        self._create_session = _make_desktop_session_provider_by_arguments(
            session_name=name,
            app_key=app_key,
        )

    def get_session(self):
        session = self._create_session()
        return session



# coding: utf8
from concurrent.futures import Future, wait
from typing import Optional, Any
from typing import TYPE_CHECKING

from ._protocol_type import ProtocolType
from ._stream_cxn_cache import stream_cxn_cache
from .event import StreamEvent, StreamCxnEvent
from .stream_state import StreamState
from .stream_state_manager import StreamStateManager
from ..._core.session.tools import is_closed

if TYPE_CHECKING:
    from ...content._content_type import ContentType
    from .._stream import StreamConnection
    from ..._core.session import Session
    from ._stream_factory import StreamDetails


class Stream(StreamStateManager):
    _cxn: Optional["StreamConnection"] = None
    _event: Optional[StreamEvent] = None

    def __init__(
        self,
        stream_id: int,
        session: "Session",
        details: "StreamDetails",
    ) -> None:
        StreamStateManager.__init__(self, logger=session.logger())
        self.details = details
        self._id: int = stream_id
        self._session: "Session" = session
        self._opened: Optional[Future] = None
        self._closed: Optional[Future] = None
        self._classname = f"[{self.__class__.__name__}_{self.id}]"

    @property
    def classname(self):
        return self._classname

    @property
    def id(self) -> int:
        return self._id

    @property
    def session(self) -> "Session":
        return self._session

    @property
    def name(self) -> str:
        return ""

    @property
    def content_type(self) -> "ContentType":
        return self.details.content_type

    @property
    def protocol_type(self) -> ProtocolType:
        return ProtocolType.NONE

    @property
    def close_message(self) -> dict:
        return {}

    @property
    def open_message(self) -> dict:
        return {}

    def open(self, *args, with_updates: bool = True) -> StreamState:
        if is_closed(self._session):
            raise AssertionError("Session must be open")

        return super().open(with_updates=with_updates)

    def send(self, message) -> None:
        self._debug(f"{self._classname} send message={message}")
        self._cxn.send(message)

    def _initialize_cxn(self):
        self._event = StreamEvent.get(self.id)
        self._cxn = stream_cxn_cache.get_cxn(self.session, self.details)
        self._cxn.on(StreamCxnEvent.DISCONNECTING, self.close)
        self._cxn.on(StreamCxnEvent.RECONNECTED, self._on_reconnected)
        self._cxn.on(StreamCxnEvent.DISPOSED, self.halt)

    def _release_cxn(self):
        self._cxn.remove_listener(StreamCxnEvent.DISCONNECTING, self.close)
        self._cxn.remove_listener(StreamCxnEvent.RECONNECTED, self._on_reconnected)
        self._cxn.remove_listener(StreamCxnEvent.DISPOSED, self.halt)
        self._debug(f"{self._classname} release cxn={self._cxn}")
        stream_cxn_cache.release(self.session, self.details)
        self._cxn = None

    def _do_open(self, *args, **kwargs) -> None:
        self._opened = Future()
        self.send(self.open_message)
        wait([self._opened])

    def _do_close(self, *args, **kwargs):
        self._closed = Future()
        self.send(self.close_message)
        self._dispose()

    def _on_reconnected(self, *args, **kwargs):
        if self.is_open:
            self.send(self.open_message)

    def _do_on_stream_error(self, originator, *args) -> Any:
        if self.is_opening:
            not self._opened.done() and self._opened.set_result(False)

        return args


def update_message_with_extended_params(message: dict, extended_params: dict) -> dict:
    return update_key_in_dict(message, extended_params)


def update_key_in_dict(message: dict, extended_params: dict) -> dict:
    for param, extended_val in extended_params.items():
        if param in message:
            prev_value = message[param]
            if isinstance(prev_value, dict) and isinstance(extended_val, dict):
                update_key_in_dict(prev_value, extended_val)
            else:
                message[param] = extended_val
        else:
            message[param] = extended_val

    return message



# coding: utf8

from ._enums import UnderlyingType
from ._models import SurfaceOutput as SurfaceLayout
from .._object_definition import ObjectDefinition


class SurfaceRequestItem(ObjectDefinition):
    def __init__(
        self,
        surface_layout,
        surface_tag,
        underlying_type,
    ):
        super().__init__()
        self.surface_tag = surface_tag
        self.surface_layout = surface_layout
        self.underlying_type = underlying_type

    @property
    def surface_layout(self):
        """
        The section that contains the properties that define how the volatility surface is returned
        :return: object SurfaceLayout
        """
        return self._get_object_parameter(SurfaceLayout, "surfaceLayout")

    @surface_layout.setter
    def surface_layout(self, value):
        self._set_object_parameter(SurfaceLayout, "surfaceLayout", value)

    @property
    def underlying_type(self):
        """
        The type of the underlying used to generate the volatility surface
        :return: enum UnderlyingType
        """
        return self._get_enum_parameter(UnderlyingType, "underlyingType")

    @underlying_type.setter
    def underlying_type(self, value):
        self._set_enum_parameter(UnderlyingType, "underlyingType", value)

    @property
    def surface_tag(self):
        """
        A user defined string to describe the volatility surface
        :return: str
        """
        return self._get_parameter("surfaceTag")

    @surface_tag.setter
    def surface_tag(self, value):
        self._set_parameter("surfaceTag", value)



# coding: utf8

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class AmericanMonteCarloMethod(Enum):
    ANDERSEN = "Andersen"
    LONGSTAFF_SCHWARTZ = "LongstaffSchwartz"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            AmericanMonteCarloMethod, _AMERICAN_MONTE_CARLO_METHOD_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _AMERICAN_MONTE_CARLO_METHOD_VALUES_IN_LOWER_BY_AMERICAN_MONTE_CARLO_METHOD,
            some,
        )


_AMERICAN_MONTE_CARLO_METHOD_VALUES = tuple(t.value for t in AmericanMonteCarloMethod)
_AMERICAN_MONTE_CARLO_METHOD_VALUES_IN_LOWER_BY_AMERICAN_MONTE_CARLO_METHOD = {
    name.lower(): item for name, item in AmericanMonteCarloMethod.__members__.items()
}



from .._content_provider import ContentProviderLayer
from .._content_type import ContentType
from ..._tools import create_repr, validate_bool_value


class Definition(ContentProviderLayer):
    """
    This class describe parameters to retrieve data for ESG universe.

    Parameters
    ----------
    closure : str, optional
        Specifies the parameter that will be merged with the request

    use_field_names_in_headers: bool, optional
        Return field name as column headers for data instead of title

    Examples
    --------
    >>> from refinitiv.data.content import esg
    >>> definition = esg.universe.Definition()
    >>> response = definition.get_data()

    >>> response = await definition.get_data_async()
    """

    def __init__(
        self,
        closure: str = None,
        use_field_names_in_headers: bool = False,
    ):
        validate_bool_value(use_field_names_in_headers)

        super().__init__(
            content_type=ContentType.ESG_UNIVERSE,
            closure=closure,
            use_field_names_in_headers=use_field_names_in_headers,
        )

    def __repr__(self):
        return create_repr(
            self,
            middle_path="universe",
            content=f"{{closure='{self._kwargs.get('closure')}'}}",
        )



# coding: utf8

__all__ = ["RiskType"]

from enum import Enum, unique

from ._common_tools import _convert_to_str, _normalize


@unique
class RiskType(Enum):
    INTEREST_RATE = "InterestRate"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(RiskType, _RISK_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_RISK_TYPE_VALUES_IN_LOWER_BY_RISK_TYPE, some)


_RISK_TYPE_VALUES = tuple(t.value for t in RiskType)
_RISK_TYPE_VALUES_IN_LOWER_BY_RISK_TYPE = {
    name.lower(): item for name, item in RiskType.__members__.items()
}



__all__ = (
    "Definition",
    "Events",
    "FinalizedOrders",
    "UniverseTypes",
)

from ._definition import Definition
from ._stream import Events
from ._stream import FinalizedOrders
from ._stream import UniverseTypes



from ..._enums import (
    AdjustInterestToPaymentDate,
    BusinessDayConvention,
    BuySell,
    DateRollingConvention,
    DayCountBasis,
    Frequency,
    IndexConvexityAdjustmentIntegrationMethod,
    IndexConvexityAdjustmentMethod,
    IndexResetType,
    StubRule,
    AmortizationType,
)



from typing import Optional, Iterable, Union, TYPE_CHECKING

from ..._types import Strings, OptStr, ExtendedParams, OptStrings, OptFloat, OptInt

if TYPE_CHECKING:
    from ._swap_zc_curve_definition import SwapZcCurveDefinition
    from ._swap_zc_curve_parameters import SwapZcCurveParameters
    from ._forward_curve_definition import ForwardCurveDefinition
    from .. import forward_curve
    from ._enums import ForwardCurvesOutputs

CurveDefinition = Optional["SwapZcCurveDefinition"]
CurveParameters = Optional["SwapZcCurveParameters"]
ForwardCurveDefinitions = Union[
    "ForwardCurveDefinition", Iterable["ForwardCurveDefinition"]
]

Outputs = Union[Strings, Iterable["ForwardCurvesOutputs"]]
Universe = Union["forward_curve.Definition", Iterable["forward_curve.Definition"]]

Steps = Union[Iterable["Step"]]
Turns = Union[Iterable["Turn"]]



# coding: utf8

__all__ = ["VolatilityTermStructureType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class VolatilityTermStructureType(Enum):
    HISTORICAL = "Historical"
    IMPLIED = "Implied"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            VolatilityTermStructureType, _VOLATILITY_TERM_STRUCTURE_TYPE_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _VOLATILITY_TERM_STRUCTURE_TYPE_VALUES_IN_LOWER_BY_VOLATILITY_TERM_STRUCTURE_TYPE,
            some,
        )


_VOLATILITY_TERM_STRUCTURE_TYPE_VALUES = tuple(
    t.value for t in VolatilityTermStructureType
)
_VOLATILITY_TERM_STRUCTURE_TYPE_VALUES_IN_LOWER_BY_VOLATILITY_TERM_STRUCTURE_TYPE = {
    name.lower(): item for name, item in VolatilityTermStructureType.__members__.items()
}



# coding: utf8

__all__ = ["OptionVolatilityType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class OptionVolatilityType(Enum):
    HISTORICAL = "Historical"
    IMPLIED = "Implied"
    SVI_SURFACE = "SVISurface"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            OptionVolatilityType, _OPTION_VOLATILITY_TYPE_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _OPTION_VOLATILITY_TYPE_VALUES_IN_LOWER_BY_OPTION_VOLATILITY_TYPE, some
        )


_OPTION_VOLATILITY_TYPE_VALUES = tuple(t.value for t in OptionVolatilityType)
_OPTION_VOLATILITY_TYPE_VALUES_IN_LOWER_BY_OPTION_VOLATILITY_TYPE = {
    name.lower(): item for name, item in OptionVolatilityType.__members__.items()
}



# coding: utf8
# contract_gen 2020-05-18 08:30:59.217815

__all__ = ["FixingFrequency"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class FixingFrequency(Enum):
    DAILY = "Daily"
    BI_WEEKLY = "BiWeekly"
    WEEKLY = "Weekly"
    MONTHLY = "Monthly"
    QUATERLY = "Quaterly"
    SEMI_ANNUAL = "SemiAnnual"
    ANNUAL = "Annual"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(FixingFrequency, _FIXING_FREQUENCY_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_FIXING_FREQUENCY_VALUES_IN_LOWER_BY_FIXING_FREQUENCY, some)


_FIXING_FREQUENCY_VALUES = tuple(t.value for t in FixingFrequency)
_FIXING_FREQUENCY_VALUES_IN_LOWER_BY_FIXING_FREQUENCY = {
    name.lower(): item for name, item in FixingFrequency.__members__.items()
}



# coding: utf-8

__all__ = ("Definition", "GrantRefreshToken", "GrantPassword")

from .._core.session.grant_refresh import GrantRefreshToken
from .._core.session.grant_password import GrantPassword


class Definition(object):
    """
    Create definition object for platform session.
    When user provides "session-name", then it is checked
    in the config file for matches, if match is found,
    then this session will be taken, otherwise, an exception will be raised

    When user doesn't provide a session name,
    then the default session from the config file is taken

    Parameters
    ----------
        name: str
            name of the session in config. For example: 'default'
        app_key: str
            application key
        grant: GrantPassword object
            GrantPassword object created based on user's password and username
        signon_control: bool
            signon_control value to take sign on control
        deployed_platform_host: str
            for Deployed platform connection.
            The host contains host name and
            port i.e. 127.0.0.0:15000
        deployed_platform_username: str
            deployed platform username
        dacs_position: str
            socket host position by name
        dacs_application_id: str
            dacs application id, default value: 256

    Raises
    ----------
    Exception
        If app-key is not found in the config file and in arguments.

    Examples
    --------
    >>> import refinitiv.data as rd
    >>> definition = rd.session.platform.Definition(name="custom-session-name")
    >>> platform_session = definition.get_session()
    """

    def __init__(
        self,
        name: str = "default",
        app_key: str = None,
        grant=None,
        signon_control: bool = None,
        deployed_platform_host: str = None,
        deployed_platform_username: str = None,
        dacs_position: str = None,
        dacs_application_id: str = None,
    ) -> None:
        from .._core.session._session_provider import (
            _make_platform_session_provider_by_arguments,
        )

        if not isinstance(name, str):
            raise ValueError("Invalid session name type, please provide string.")

        self._create_session = _make_platform_session_provider_by_arguments(
            session_name=name,
            app_key=app_key,
            signon_control=signon_control,
            deployed_platform_host=deployed_platform_host,
            deployed_platform_username=deployed_platform_username,
            dacs_position=dacs_position,
            dacs_application_id=dacs_application_id,
            grant=grant,
        )

    def get_session(self):
        session = self._create_session()
        return session



from enum import Enum, unique, auto


@unique
class StreamCxnState(Enum):
    Initial = auto()
    Connecting = auto()
    Connected = auto()
    Ready = auto()
    Disconnecting = auto()
    Disconnected = auto()
    Reconnecting = auto()
    Disposed = auto()



from numpy import iterable

from .._ipa_content_provider import IPAContentProviderLayer
from ..._content_type import ContentType


class Definition(IPAContentProviderLayer):
    def __init__(
        self,
        universe,
        surface_layout=None,
        extended_params=None,
    ):
        if not iterable(universe):
            universe = [universe]

        super().__init__(
            content_type=ContentType.SURFACES,
            universe=universe,
            extended_params=extended_params,
        )



# coding: utf-8

from typing import TYPE_CHECKING

from .._content_provider import ContentProviderLayer
from ..._tools import create_repr

if TYPE_CHECKING:
    from ._search_view_type import SearchViews


class Definition(ContentProviderLayer):
    """
    This class describe parameters to retrieve data for search metadata.

    Parameters
    ----------

    view : SearchViews
        picks a subset of the data universe to search against. see SearchViews

    Examples
    --------
    >>> from refinitiv.data.content import search
    >>> definition = search.metadata.Definition(view=search.SearchViews.PEOPLE)
    """

    def __init__(self, view: "SearchViews"):
        self._view = view

        from .._content_type import ContentType

        super().__init__(
            content_type=ContentType.DISCOVERY_METADATA,
            view=self._view,
        )

    def __repr__(self):
        return create_repr(
            self,
            middle_path="metadata",
            content=f"{{view='{self._view}'}}",
        )



from typing import Callable, Any, Dict, Union

import pandas as pd

from ._content_type import ContentType
from ._df_build_type import DFBuildType
from ._df_builder import (
    dfbuilder_udf,
    dfbuilder_rdp,
    default_build_df,
    build_empty_df,
    historical_build_df,
)
from .custom_instruments._data_provider import (
    custom_instruments_intervals_build_df,
    custom_instruments_build_df,
)
from .ipa._curves._curves_data_provider import (
    forward_curve_build_df,
    zc_curve_definitions_build_df,
    zc_curves_build_df,
)
from .ipa.financial_contracts._contracts_data_provider import (
    financial_contracts_build_df,
)
from .news._tools import news_build_df_rdp, news_build_df_udf
from .pricing._pricing_content_provider import pricing_build_df
from .pricing.chain._chains_data_provider import chains_build_df
from .search._data_provider import (
    discovery_lookup_build_df,
    discovery_metadata_build_df,
    discovery_search_build_df,
)
from ..delivery._data._data_type import DataType
from ..delivery.cfs._cfs_data_provider import cfs_build_df

content_type_by_build_type: Dict[
    Union[ContentType, DataType], Callable[[Any, Dict[str, Any]], pd.DataFrame]
] = {
    ContentType.CHAINS: chains_build_df,
    ContentType.CONTRACTS: financial_contracts_build_df,
    ContentType.CUSTOM_INSTRUMENTS_EVENTS: historical_build_df,
    ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS: custom_instruments_build_df,
    ContentType.CUSTOM_INSTRUMENTS_INTERDAY_SUMMARIES: historical_build_df,
    ContentType.CUSTOM_INSTRUMENTS_INTRADAY_SUMMARIES: historical_build_df,
    ContentType.CUSTOM_INSTRUMENTS_SEARCH: custom_instruments_build_df,
    ContentType.DATA_GRID_RDP: {
        DFBuildType.INDEX: dfbuilder_rdp.build_index,
        DFBuildType.DATE_AS_INDEX: dfbuilder_rdp.build_date_as_index,
    },
    ContentType.DATA_GRID_UDF: {
        DFBuildType.INDEX: dfbuilder_udf.build_index,
        DFBuildType.DATE_AS_INDEX: dfbuilder_udf.build_date_as_index,
    },
    ContentType.DEFAULT: default_build_df,
    ContentType.DISCOVERY_LOOKUP: discovery_lookup_build_df,
    ContentType.DISCOVERY_METADATA: discovery_metadata_build_df,
    ContentType.DISCOVERY_SEARCH: discovery_search_build_df,
    ContentType.ESG_BASIC_OVERVIEW: dfbuilder_rdp.build_index,
    ContentType.ESG_FULL_MEASURES: dfbuilder_rdp.build_index,
    ContentType.ESG_FULL_SCORES: dfbuilder_rdp.build_index,
    ContentType.ESG_STANDARD_MEASURES: dfbuilder_rdp.build_index,
    ContentType.ESG_STANDARD_SCORES: dfbuilder_rdp.build_index,
    ContentType.ESG_UNIVERSE: dfbuilder_rdp.build_index,
    ContentType.ESTIMATES_VIEW_ACTUALS_ANNUAL: dfbuilder_rdp.build_index,
    ContentType.ESTIMATES_VIEW_ACTUALS_INTERIM: dfbuilder_rdp.build_index,
    ContentType.ESTIMATES_VIEW_ACTUALS_KPI_ANNUAL: dfbuilder_rdp.build_index,
    ContentType.ESTIMATES_VIEW_ACTUALS_KPI_INTERIM: dfbuilder_rdp.build_index,
    ContentType.ESTIMATES_VIEW_SUMMARY_ANNUAL: dfbuilder_rdp.build_index,
    ContentType.ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_NON_PERIODIC_MEASURES: dfbuilder_rdp.build_index,
    ContentType.ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_PERIODIC_MEASURES_ANNUAL: dfbuilder_rdp.build_index,
    ContentType.ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_PERIODIC_MEASURES_INTERIM: dfbuilder_rdp.build_index,
    ContentType.ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_RECOMMENDATIONS: dfbuilder_rdp.build_index,
    ContentType.ESTIMATES_VIEW_SUMMARY_INTERIM: dfbuilder_rdp.build_index,
    ContentType.ESTIMATES_VIEW_SUMMARY_KPI_ANNUAL: dfbuilder_rdp.build_index,
    ContentType.ESTIMATES_VIEW_SUMMARY_KPI_HISTORICAL_SNAPSHOTS_KPI: dfbuilder_rdp.build_index,
    ContentType.ESTIMATES_VIEW_SUMMARY_KPI_INTERIM: dfbuilder_rdp.build_index,
    ContentType.ESTIMATES_VIEW_SUMMARY_NON_PERIODIC_MEASURES: dfbuilder_rdp.build_index,
    ContentType.ESTIMATES_VIEW_SUMMARY_RECOMMENDATIONS: dfbuilder_rdp.build_index,
    ContentType.FILINGS_RETRIEVAL: dfbuilder_rdp.build_index,
    ContentType.FORWARD_CURVE: forward_curve_build_df,
    ContentType.HISTORICAL_PRICING_EVENTS: historical_build_df,
    ContentType.HISTORICAL_PRICING_INTERDAY_SUMMARIES: historical_build_df,
    ContentType.HISTORICAL_PRICING_INTRADAY_SUMMARIES: historical_build_df,
    ContentType.NEWS_HEADLINES_RDP: news_build_df_rdp,
    ContentType.NEWS_HEADLINES_UDF: news_build_df_udf,
    ContentType.NEWS_STORY_RDP: build_empty_df,
    ContentType.NEWS_STORY_UDF: build_empty_df,
    ContentType.OWNERSHIP_CONSOLIDATED_BREAKDOWN: dfbuilder_rdp.build_index,
    ContentType.OWNERSHIP_CONSOLIDATED_CONCENTRATION: dfbuilder_rdp.build_index,
    ContentType.OWNERSHIP_CONSOLIDATED_INVESTORS: dfbuilder_rdp.build_index,
    ContentType.OWNERSHIP_CONSOLIDATED_RECENT_ACTIVITY: dfbuilder_rdp.build_index,
    ContentType.OWNERSHIP_CONSOLIDATED_SHAREHOLDERS_HISTORY_REPORT: dfbuilder_rdp.build_index,
    ContentType.OWNERSHIP_CONSOLIDATED_SHAREHOLDERS_REPORT: dfbuilder_rdp.build_index,
    ContentType.OWNERSHIP_CONSOLIDATED_TOP_N_CONCENTRATION: dfbuilder_rdp.build_index,
    ContentType.OWNERSHIP_FUND_BREAKDOWN: dfbuilder_rdp.build_index,
    ContentType.OWNERSHIP_FUND_CONCENTRATION: dfbuilder_rdp.build_index,
    ContentType.OWNERSHIP_FUND_HOLDINGS: dfbuilder_rdp.build_index,
    ContentType.OWNERSHIP_FUND_INVESTORS: dfbuilder_rdp.build_index,
    ContentType.OWNERSHIP_FUND_RECENT_ACTIVITY: dfbuilder_rdp.build_index,
    ContentType.OWNERSHIP_FUND_SHAREHOLDERS_HISTORY_REPORT: dfbuilder_rdp.build_index,
    ContentType.OWNERSHIP_FUND_SHAREHOLDERS_REPORT: dfbuilder_rdp.build_index,
    ContentType.OWNERSHIP_FUND_TOP_N_CONCENTRATION: dfbuilder_rdp.build_index,
    ContentType.OWNERSHIP_INSIDER_SHAREHOLDERS_REPORT: dfbuilder_rdp.build_index,
    ContentType.OWNERSHIP_INSIDER_TRANSACTION_REPORT: dfbuilder_rdp.build_index,
    ContentType.OWNERSHIP_INVESTOR_HOLDINGS: dfbuilder_rdp.build_index,
    ContentType.OWNERSHIP_ORG_INFO: dfbuilder_rdp.build_index,
    ContentType.PRICING: pricing_build_df,
    ContentType.SURFACES: default_build_df,
    ContentType.ZC_CURVE_DEFINITIONS: zc_curve_definitions_build_df,
    ContentType.ZC_CURVES: zc_curves_build_df,
    DataType.CFS_BUCKETS: cfs_build_df,
    DataType.CFS_FILE_SETS: cfs_build_df,
    DataType.CFS_FILES: cfs_build_df,
    DataType.CFS_PACKAGES: cfs_build_df,
    DataType.CFS_STREAM: default_build_df,
    DataType.ENDPOINT: default_build_df,
}


def get_dfbuilder(
    content_type: ContentType, dfbuild_type: DFBuildType = DFBuildType.INDEX
) -> Callable[[Any, Dict[str, Any]], pd.DataFrame]:
    builder_by_build_type = content_type_by_build_type.get(content_type)

    if not builder_by_build_type:
        raise ValueError(f"Cannot find mapping for content_type={content_type}")

    if hasattr(builder_by_build_type, "get"):
        builder = builder_by_build_type.get(dfbuild_type)

    else:
        builder = builder_by_build_type

    if not builder:
        raise ValueError(f"Cannot find mapping for dfbuild_type={dfbuild_type}")

    return builder



import json
import os
import os.path

from ._actions import Actions
from ._errors import PackageNotFoundException
from ._tools import (
    get_date_from_filename,
    get_the_newest_file,
    get_init_archives,
    get_filesets_with_delta_files,
    get_filesets_with_the_newest_init_file_and_delta_files,
    sorted_files,
)
from ...._core.log_reporter import _LogReporter
from ....delivery.cfs._tools import path_join, get_correct_filename, remove_one_ext
from ....delivery.cfs._unpacker import _Unpacker
from ....delivery import cfs


class FileManager(_LogReporter):
    def __init__(
        self,
        logger,
        actions: Actions,
        package_name,
        bucket,
        path,
        session=None,
        auto_retry=0,
        auto_extract=False,
    ):
        super().__init__(logger=logger)
        self._package_name = package_name
        self._bucket = bucket
        self._path = path
        self._auto_extract = auto_extract
        self._auto_retry_count = auto_retry
        self._actions = actions
        self._unpacker = _Unpacker(logger)
        self.session = session

    def update_files(self):
        _packages_def = cfs.packages.Definition(
            package_name=self._package_name,
            bucket_name=self._bucket,
        )
        _packages = _packages_def.get_data(session=self.session)

        if not _packages.data.packages:
            message = (
                f"package (name={self._package_name}, bucket={self._bucket}) not found"
            )
            self._actions.add(
                "PACKAGE NOT FOUND",
                bucket=self._bucket,
                package=self._package_name,
            )
            self._info(
                f"package "
                f"bucket='{self._bucket}', package_name='{self._package_name}' "
                f"not found"
            )
            raise PackageNotFoundException(message)

        _package_id = _packages.data.packages[0].package_id

        self._actions.update()
        downloaded = self._actions.get_downloaded()

        init_files = get_init_archives(downloaded) if downloaded else None
        init_file = get_the_newest_file(init_files) if init_files else None

        init_file_date = get_date_from_filename(init_file) if init_file else None

        # if init_file_date is None, the modified_since parameter will be skipped
        _file_sets_def = cfs.file_sets.Definition(
            bucket=self._bucket,
            package_id=_package_id,
            modified_since=init_file_date,
        )
        _file_sets = _file_sets_def.get_data(session=self.session)
        _file_sets = [file_set for file_set in _file_sets.data.file_sets]

        if not init_file:
            # if no init files are loaded,
            # get the latest init and delta files with the same or newer date
            _file_sets = get_filesets_with_the_newest_init_file_and_delta_files(
                _file_sets
            )

        else:
            # if init file already downloaded, get deltas with same or newer date
            _file_sets = get_filesets_with_delta_files(_file_sets)

        # download files
        has_files_for_download = False
        for file_set in _file_sets:
            for file in file_set:
                filename_ext = get_correct_filename(file.filename)
                if filename_ext not in downloaded:
                    has_files_for_download = True
                    self.__retry_download_file(file.id, filename_ext)
        if not has_files_for_download:
            self._info("no new files for download")

    def __retry_download_file(self, file_id: str, filename_ext: str):
        count_of_try = self._auto_retry_count + 1

        while count_of_try:
            try:
                self._info(f"download file {filename_ext} to {self._path}")
                fd = cfs.file_downloader.Definition(
                    {"id": file_id, "filename": filename_ext}
                ).retrieve(session=self.session)
                fd.download(self._path)
                self._actions.downloaded(path=self._path, filename=filename_ext)

                if self._auto_extract:
                    fd.extract(self._path)
                    extracted_filename = remove_one_ext(filename_ext)
                    self._actions.extracted(
                        path=self._path, filename=extracted_filename
                    )
                    self._info(f"extract file {filename_ext} to {self._path}")

                break
            except Exception as e:
                details = self._actions.add(
                    "DOWNLOAD CRASHED",
                    path=self._path,
                    filename=filename_ext,
                    error=str(e),
                )
                self._info(f"download crashed {details}")
                count_of_try -= 1

    def extract_file(self, filename):
        self._unpacker.unpack(filename, self._path)
        self._actions.extracted(filename=filename, path=self._path)
        self._info(f"extract file '{filename}' to '{self._path}'")

    def read_file(self, filename):
        data = []
        filepath = path_join(self._path, filename)
        if not os.path.exists(filepath):
            message = f"cannot update database, because previously downloaded file is missing: {filepath}"
            self._actions.add("ERROR", message=message)
            self._error(f"error: {message}")
            raise FileNotFoundError(message)
        with open(filepath) as f:
            for line_n, line in enumerate(f):
                try:
                    data.append(json.loads(line))
                except Exception as e:
                    error_details = self._actions.add(
                        "ERROR", message=str(e), line_number=line_n, filename=filename
                    )
                    self._info(f"error: {error_details}")
                    continue

        return data

    def read_files_for_update(self) -> list:
        self._actions.update()
        files_to_update, extracted = self._actions.get_not_updated_and_extracted()

        if not files_to_update:
            return []

        data = []
        files_to_update = sorted_files(files_to_update)
        for filename in files_to_update:
            if filename not in extracted:
                self.extract_file(filename)

            file_attrs = {"DateTime": get_date_from_filename(filename)}
            data.append((filename, self.read_file(filename), file_attrs))

        return data

    def cleanup_files(self):
        # example: filenames = {"log.txt"}
        filenames = set()
        self._actions.update()
        filenames.update(self._actions.get_downloaded())
        filenames.update(self._actions.get_extracted())

        handlers = self._actions._file_logger.handlers
        for handler in handlers:
            handler.close()

        if filenames:
            filenames.update({"log.txt"})

        for filename in filenames:
            path = path_join(self._path, filename)

            try:
                os.remove(path)
            except PermissionError:
                if filename == "log.txt":
                    with open(path, "w"):
                        # only create a new file
                        pass
                    self._info(
                        f"WARNING: Cannot remove {path}. The log has been cleared."
                    )
                self._info(f"WARNING: Cannot remove {path}")



"""Module to create and use templated search presets for the discovery search

Use templated search from the "Equity" template that was defined in the config:
>>> import refinitiv.data as rd
>>> rd.content.search._templates["Equity"].search()
"""
__all__ = ["templates"]

import operator
import string
import inspect
from functools import reduce
from typing import Iterable, Optional, Dict, Set, List, Any

from humps import depascalize

from refinitiv.data import get_config
from refinitiv.data.errors import ConfigurationError
from refinitiv.data.content.search._definition import Definition
from refinitiv.data.content.search._search_view_type import SearchViews

CONFIG_PREFIX = "search.templates"


def join_sets(sets: Iterable[set]) -> set:
    """Join multiple sets into one"""
    iterator = iter(sets)
    try:
        # Can't pass an empty iterator to `reduce`, that's the way to handle it
        default = next(iterator)
    except StopIteration:
        return set()
    else:
        return reduce(operator.or_, iterator, default)


def get_placeholders_names(template_string: str) -> Set[str]:
    """Get list of placeholder names in format-style string

    >>> get_placeholders_names("{a}, {b}, {a}")
    {'b', 'a'}
    """
    return set(x[1] for x in string.Formatter().parse(template_string) if x[1])


class SearchTemplate:
    """Discovery search preset

    Initialized with default values for content.search.Definition.
    Any string value acts as template string. You can use placeholder variables
    like for `str.format` method, and that variables will be required to prepare
    search parameters through `._search_kwargs()` or to launch search through `.search()`.

    Attributes
    ----------

    name: str
        name of the template

    Examples
    --------

    >>> SearchTemplate(filter="ExchangeName xeq '{name}'")._search_kwargs(name="<name>")
    {'filter': "ExchangeName xeq '<name>'"}
    >>> SearchTemplate(filter="ExchangeName xeq '{name}'")._search_kwargs()
    Traceback (most recent call last):
        ...
    KeyError: 'Those keyword arguments must be defined, but they are not: name'
    >>> SearchTemplate(filter="ExchangeName xeq '{name}'", placeholders_defaults={"name": "<name>"})._search_kwargs()
    {'filter': "ExchangeName xeq '<name>'"}
    """

    def __init__(
        self,
        name=None,
        placeholders_defaults: Optional[Dict[str, Any]] = None,
        **search_defaults,
    ):
        """
        Parameters
        ----------
        name : str, optional
            name of the template
        placeholders_defaults: dict, optional
            <placeholder_name> : <placeholder_default_value>
        search_defaults
            default values for discovery search Definition
        """
        # List search keyword arguments we can use in this template
        self._available_search_kwargs = inspect.signature(Definition).parameters.keys()
        # Default template variables values for a templated defaults
        self._placeholders_defaults = (
            {} if placeholders_defaults is None else placeholders_defaults
        )
        self.name = name

        unknown_defaults = set(search_defaults) - set(self._available_search_kwargs)
        if unknown_defaults:
            raise ValueError(
                "This arguments are defined in template, but not in search Definition: "
                + ", ".join(unknown_defaults)
            )

        # Set of placeholders names for each templated default
        self._placeholders_vars: Dict[str, Set[str]] = {
            name: get_placeholders_names(value)
            for name, value in search_defaults.items()
            if isinstance(value, str)
        }

        self._templated_defaults = {
            name: value
            for name, value in search_defaults.items()
            if self._placeholders_vars.get(name)
        }
        self._other_defaults = {
            name: value
            for name, value in search_defaults.items()
            if name not in self._templated_defaults
        }

        bad_tpl_var_names = (
            join_sets(self._placeholders_vars.values()) & self._available_search_kwargs
        )
        if bad_tpl_var_names:
            raise ValueError(
                "You can't use template arguments with the same name"
                " as search arguments. You are used: " + ", ".join(bad_tpl_var_names)
            )

    def __repr__(self):
        return f"<SearchTemplate '{self.name}'>"

    def _search_kwargs(self, **kwargs) -> dict:
        """Get dictionary of arguments for content.search.Definition"""

        non_redefined_templated_defaults = set(self._templated_defaults) - set(kwargs)
        current_templates = [
            name
            for name, value in self._templated_defaults.items()
            if name in non_redefined_templated_defaults and isinstance(value, str)
        ]

        template_vars_names = join_sets(
            var_names
            for template_name, var_names in self._placeholders_vars.items()
            if template_name in current_templates
        )

        undefined_placeholders = (
            template_vars_names - set(kwargs) - set(self._placeholders_defaults)
        )

        if undefined_placeholders:
            raise KeyError(
                "Those keyword arguments must be defined, but they are not: "
                + ", ".join(undefined_placeholders)
            )

        unexpected_arguments = (
            set(kwargs) - template_vars_names - self._available_search_kwargs
        )

        if unexpected_arguments:
            raise KeyError(f"Unexpected arguments: {', '.join(unexpected_arguments)}")

        kwargs = kwargs.copy()
        # Applying template variables defaults
        for name, value in self._placeholders_defaults.items():
            if name not in kwargs:
                kwargs[name] = value

        result = self._other_defaults.copy()

        # Apply variables to templated defaults
        for name in current_templates:
            result[name] = self._templated_defaults[name].format(**kwargs)

        # Apply other variables from kwargs
        for name, value in kwargs.items():
            if name not in template_vars_names:
                result[name] = value

        return result

    def search(self, **kwargs):
        return Definition(**self._search_kwargs(**kwargs)).get_data().data.df


class Templates:
    """Easy access to search templates from the library config

    Check if search template with the name "Equity" is defined in the config:
    >>> templates = Templates()
    >>> "Equity" in templates
    True
    Get "Equity" search template:
    >>> templates["Equity"]
    Get list of available search template names:
    >>> templates.keys()
    ["Equity"]
    """

    def __iter__(self):
        config = get_config()
        return config.get(CONFIG_PREFIX, {}).keys().__iter__()

    def __getitem__(self, name: str) -> SearchTemplate:
        config = get_config()
        key = f"{CONFIG_PREFIX}.{name}"
        if key not in config:
            raise KeyError(f"Template '{name}' is not found in the config")
        data = config[key].as_attrdict() if config[key] is not None else {}

        # <param_name>: {"default": <default>, "description": <doc>}
        tpl_strs_defaults = {
            name: attrs["default"]
            for name, attrs in data.get("parameters", {}).items()
            if "default" in attrs
        }
        params = depascalize(data.get("request_body", {}))

        if "view" in params:
            # Convert string value to enum for the view argument
            view = params["view"]
            try:
                params["view"] = getattr(SearchViews, depascalize(view).upper())
            except AttributeError:
                raise ConfigurationError(
                    -1,
                    f"Wrong search template value: View={view}. "
                    "It must be one of the following: "
                    f"{', '.join(SearchViews.possible_values())}",
                )
        return SearchTemplate(name, placeholders_defaults=tpl_strs_defaults, **params)

    def keys(self) -> List[str]:
        """Get list of available search template names"""
        return list(self)


templates = Templates()



from ._data_provider import (
    DataProvider,
    ContentValidator,
    ResponseFactory,
    ValidatorContainer,
    RequestFactory,
)
from ._endpoint_data import EndpointData


class EndpointValidator(ContentValidator):
    def validate(self, data, *args, **kwargs):
        return True


class EndpointRequestFactory(RequestFactory):
    def extend_body_parameters(self, body_parameters, extended_params):
        return body_parameters


endpoint_data_provider = DataProvider(
    validator=ValidatorContainer(content_validator=EndpointValidator()),
    response=ResponseFactory(data_class=EndpointData),
    request=EndpointRequestFactory(),
)



# coding: utf8

from enum import Enum, unique

from ..._enums._common_tools import _convert_to_str, _normalize


@unique
class ZcCurvesOutputs(Enum):
    CONSTITUENTS = "Constituents"
    DETAILED_CURVE_POINT = "DetailedCurvePoint"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(ZcCurvesOutputs, _ZC_CURVES_OUTPUTS_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_ZC_CURVES_OUTPUTS_VALUES_IN_LOWER_BY_ZC_CURVES_OUTPUTS, some)


_ZC_CURVES_OUTPUTS_VALUES = tuple(t.value for t in ZcCurvesOutputs)
_ZC_CURVES_OUTPUTS_VALUES_IN_LOWER_BY_ZC_CURVES_OUTPUTS = {
    name.lower(): item for name, item in ZcCurvesOutputs.__members__.items()
}



# coding: utf8

__all__ = ["DateRollingConvention"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class DateRollingConvention(Enum):
    """
    Method to adjust payment dates when they fall at the end of the month (28th of
    February, 30th, 31st).

    The possible values are:
        - Last (For setting the calculated date to the last working day),
        - Same (For setting the calculated date to the same day . In this latter case,
            the date may be moved according to the date moving convention if it is
            a non-working day),
        - Last28 (For setting the calculated date to the last working day. 28FEB being
            always considered as the last working day),
        - Same28 (For setting the calculated date to the same day .28FEB being always
            considered as the last working day).
    """

    LAST = "Last"
    LAST28 = "Last28"
    SAME = "Same"
    SAME28 = "Same28"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            DateRollingConvention, _PAYMENT_ROLL_CONVENTION_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _PAYMENT_ROLL_CONVENTION_VALUES_IN_LOWER_BY_PAYMENT_ROLL_CONVENTION, some
        )


_PAYMENT_ROLL_CONVENTION_VALUES = tuple(t.value for t in DateRollingConvention)
_PAYMENT_ROLL_CONVENTION_VALUES_IN_LOWER_BY_PAYMENT_ROLL_CONVENTION = {
    name.lower(): item for name, item in DateRollingConvention.__members__.items()
}



from enum import Enum, unique

from ._tools import _normalize


@unique
class SymbolTypes(Enum):
    """
    Symbol types to send in request, by default "RIC" is using.
    """

    RIC = "RIC"
    ISIN = "IssueISIN"
    CUSIP = "CUSIP"
    SEDOL = "SEDOL"
    TICKER_SYMBOL = "TickerSymbol"
    OA_PERM_ID = "IssuerOAPermID"
    LIPPER_ID = "FundClassLipperID"

    @staticmethod
    def convert_to_str(some):
        result = None
        if isinstance(some, str):
            if some in SYMBOL_TYPE_VALUES:
                result = some
            else:
                result = SymbolTypes.normalize(some)
        elif isinstance(some, SymbolTypes):
            result = some.value
        if result:
            return result
        else:
            raise AttributeError(f"Symbol type value must be in {SYMBOL_TYPE_VALUES}")

    @staticmethod
    def normalize(some: str) -> str:
        return _normalize(some, _SYMBOL_TYPE_VALUES_IN_LOWER_BY_SYMBOL_TYPE)


SYMBOL_TYPE_VALUES = tuple(t.value for t in SymbolTypes)
_SYMBOL_TYPE_VALUES_IN_LOWER_BY_SYMBOL_TYPE = {
    name.lower(): item for name, item in SymbolTypes.__members__.items()
}



import enum


@enum.unique
class SearchViews(enum.Enum):
    """Possible views values to request data from 'search' endpoint"""

    BOND_FUT_OPT_QUOTES = "BondFutOptQuotes"
    CDS_INSTRUMENTS = "CdsInstruments"
    CDS_QUOTES = "CdsQuotes"
    CMO_INSTRUMENTS = "CmoInstruments"
    CMO_QUOTES = "CmoQuotes"
    COMMODITY_QUOTES = "CommodityQuotes"
    DEALS_MERGERS_AND_ACQUISITIONS = "DealsMergersAndAcquisitions"
    DERIVATIVE_INSTRUMENTS = "DerivativeInstruments"
    DERIVATIVE_QUOTES = "DerivativeQuotes"
    EQUITY_DERIVATIVE_INSTRUMENTS = "EquityDerivativeInstruments"
    EQUITY_DERIVATIVE_QUOTES = "EquityDerivativeQuotes"
    EQUITY_INSTRUMENTS = "EquityInstruments"
    EQUITY_QUOTES = "EquityQuotes"
    FIXED_INCOME_INSTRUMENTS = "FixedIncomeInstruments"
    FIXED_INCOME_QUOTES = "FixedIncomeQuotes"
    FUND_QUOTES = "FundQuotes"
    GOV_CORP_INSTRUMENTS = "GovCorpInstruments"
    GOV_CORP_QUOTES = "GovCorpQuotes"
    INDEX_INSTRUMENTS = "IndexInstruments"
    INDEX_QUOTES = "IndexQuotes"
    INDICATOR_QUOTES = "IndicatorQuotes"
    INSTRUMENTS = "Instruments"
    IRD_QUOTES = "IRDQuotes"
    LOAN_INSTRUMENTS = "LoanInstruments"
    LOAN_QUOTES = "LoanQuotes"
    MONEY_QUOTES = "MoneyQuotes"
    MORTGAGE_INSTRUMENTS = "MortgageInstruments"
    MORT_QUOTES = "MortQuotes"
    MUNICIPAL_INSTRUMENTS = "MunicipalInstruments"
    MUNICIPAL_QUOTES = "MunicipalQuotes"
    ORGANISATIONS = "Organisations"
    PEOPLE = "People"
    PHYSICAL_ASSETS = "PhysicalAssets"
    QUOTES = "Quotes"
    QUOTES_AND_STIRS = "QuotesAndSTIRs"
    SEARCH_ALL = "SearchAll"
    STIRS = "STIRs"
    VESSEL_PHYSICAL_ASSETS = "VesselPhysicalAssets"
    YIELD_CURVE_CONT_QUOTES = "YieldCurveContQuotes"

    @classmethod
    def possible_values(cls):
        return [s.value for s in cls.__members__.values()]



# coding: utf8

import abc

from .._object_definition import ObjectDefinition


class InstrumentDefinition(ObjectDefinition, abc.ABC):
    """
    This class is designed to represent
    the instrument definition templates for QPS request.
    """

    @classmethod
    def get_instrument_type(cls):
        return ""

    def __init__(self, instrument_tag=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instrument_tag = instrument_tag

    @property
    def instrument_tag(self):
        """
        User defined string to identify the instrument.It can be used to link output
        results to the instrument definition. Only alphabetic, numeric and '- _.#=@'
        characters are supported. Optional.
        :return: str
        """
        return self._get_parameter("instrumentTag")

    @instrument_tag.setter
    def instrument_tag(self, value):
        try:
            value = value.value
        except Exception:
            # silently
            pass
        self._set_parameter("instrumentTag", value)



from .._content_provider import (
    UniverseContentValidator,
    ErrorParser,
)
from ..._tools import universe_arg_parser
from ...delivery._data._data_provider import (
    DataProvider,
    RequestFactory,
    ValidatorContainer,
)


# ---------------------------------------------------------------------------
#   Request
# ---------------------------------------------------------------------------


class ESGRequestFactory(RequestFactory):
    def get_query_parameters(self, *args, **kwargs):
        query_parameters = []

        #
        # universe
        #
        universe = kwargs.get("universe")
        if universe:
            universe = universe_arg_parser.get_str(universe, delim=",")
            query_parameters.append(("universe", universe))

        #
        # start
        #
        start = kwargs.get("start")
        if start is not None:
            query_parameters.append(("start", start))

        #
        # end
        #
        end = kwargs.get("end")
        if end is not None:
            query_parameters.append(("end", end))

        return query_parameters


# ---------------------------------------------------------------------------
#   Provider
# ---------------------------------------------------------------------------

esg_data_provider = DataProvider(
    request=ESGRequestFactory(),
    validator=ValidatorContainer(content_validator=UniverseContentValidator()),
    parser=ErrorParser(),
)



from typing import Callable, List, Optional, Union, TYPE_CHECKING

import pandas as pd
from .get_data import _send_request
from ..content import fundamental_and_reference

from ._pricing_recorder import PricingRecorder
from ..content.pricing._stream_facade import Stream
from ..content._streamingprices import StreamingPrice
from .._core.session import get_default
from .._tools import cached_property

if TYPE_CHECKING:
    from .. import OpenState


def make_callback(on_data, logger):
    def callback(update, ric, stream):
        try:
            stream = PricingStream(stream)
            df = pd.DataFrame(update, index=[ric])
            on_data(df, ric, stream)
        except Exception as error:
            logger.error(error)

    return callback


def open_pricing_stream(
    universe: Union[str, List[str]],
    fields: Union[str, List[str]] = None,
    service: Optional[str] = None,
    on_data: Optional[Callable] = None,
) -> "PricingStream":
    """
    Create and open pricing stream

    Parameters
    ----------

    universe : str or list of str
        The single/multiple instrument/s name (e.g. "EUR=" or ["EUR=", "CAD=", "UAH="])
    fields : list, str, optional
        Specifies the specific fields to be delivered when messages arrive
    service : str, optional
        Name of the streaming service publishing the instruments
    on_data : function, optional
        Callback function for on_update and on_refresh events

    Returns
    ----------
    PricingStream

    Examples
    -------
    Create pricing stream

    >>> import refinitiv.data as rd
    >>> def callback(updated_data, ric, stream):
    ...    print(updated_data)
    >>> pricing_stream = rd.open_pricing_stream(universe=['EUR='], fields=['BID', 'ASK', 'OPEN_PRC'], on_data=callback)  # noqa
    """
    logger = get_default().logger()

    _adc_df = _send_request(
        data_provider=fundamental_and_reference.Definition,
        params={"universe": universe, "fields": ["TR.RIC"]},
        logger=logger,
    )

    universe = list(_adc_df.get("Instrument", universe))

    _stream = Stream(universe=universe, fields=fields, service=service)

    if on_data:
        _stream.on_update(make_callback(on_data, logger))
        _stream.on_refresh(make_callback(on_data, logger))
    stream = PricingStream(_stream)

    stream.open()
    return stream


class PricingStream:
    def __init__(self, stream):
        self._stream = stream

    def open(self, with_updates: bool = True) -> "OpenState":
        return self._stream.open(with_updates=with_updates)

    def close(self) -> "OpenState":
        return self._stream.close()

    def get_snapshot(
        self,
        universe: Union[str, List[str], None] = None,
        fields: Optional[List[str]] = None,
        convert: bool = True,
    ) -> "pd.DataFrame":
        return self._stream.get_snapshot(
            universe=universe, fields=fields, convert=convert
        )

    def _get_fields(self, universe: str, fields: Optional[list] = None) -> dict:
        return self._stream._get_fields(universe=universe, fields=fields)

    def __getitem__(self, item) -> "StreamingPrice":
        return self._stream.__getitem__(item)

    def __iter__(self):
        return self._stream.__iter__()

    @cached_property
    def recorder(self) -> PricingRecorder:
        return PricingRecorder(self._stream._stream)



# coding: utf8

import abc
from .._object_definition import ObjectDefinition


class InstrumentPricingParameters(ObjectDefinition, abc.ABC):
    class Params(abc.ABC):
        def __init__(self):
            self._pricing_parameters = {}

        def _with_key_parameter(self, key_name, value):
            if value:
                self._pricing_parameters[key_name] = value
            elif self._pricing_parameters.get(key_name):
                self._pricing_parameters.pop(key_name)

        @property
        def parameters(self):
            return self._pricing_parameters

    def __init__(self):
        super().__init__()



# coding: utf8

from typing import Optional

from .._base import BinaryDefinition
from .._enums import (
    BinaryType,
    UpOrDown,
)


class EtiBinaryDefinition(BinaryDefinition):
    """
    Parameters
    ----------
    notional_amount : float, optional

    binary_type : BinaryType, optional

    up_or_down : UpOrDown, optional

    level : float, optional

    """

    def __init__(
        self,
        notional_amount: Optional[float] = None,
        binary_type: Optional[BinaryType] = None,
        up_or_down: Optional[UpOrDown] = None,
        level: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.notional_amount = notional_amount
        self.binary_type = binary_type
        self.up_or_down = up_or_down
        self.level = level

    @property
    def binary_type(self):
        """
        :return: enum BinaryType
        """
        return self._get_enum_parameter(BinaryType, "binaryType")

    @binary_type.setter
    def binary_type(self, value):
        self._set_enum_parameter(BinaryType, "binaryType", value)

    @property
    def up_or_down(self):
        """
        :return: enum UpOrDown
        """
        return self._get_enum_parameter(UpOrDown, "upOrDown")

    @up_or_down.setter
    def up_or_down(self, value):
        self._set_enum_parameter(UpOrDown, "upOrDown", value)

    @property
    def level(self):
        """
        :return: float
        """
        return self._get_parameter("level")

    @level.setter
    def level(self, value):
        self._set_parameter("level", value)

    @property
    def notional_amount(self):
        """
        :return: float
        """
        return self._get_parameter("notionalAmount")

    @notional_amount.setter
    def notional_amount(self, value):
        self._set_parameter("notionalAmount", value)



# coding: utf8

from ..._object_definition import ObjectDefinition


class RepoParameters(ObjectDefinition):
    def __init__(
        self,
        coupon_paid_at_horizon=None,
        haircut_rate_percent=None,
        initial_margin_percent=None,
    ):
        super().__init__()
        self.coupon_paid_at_horizon = coupon_paid_at_horizon
        self.haircut_rate_percent = haircut_rate_percent
        self.initial_margin_percent = initial_margin_percent

    @property
    def coupon_paid_at_horizon(self):
        """
        Flag that defines wether couponis paid at horizon.  This has no impact on pricing.
        :return: bool
        """
        return self._get_parameter("couponPaidAtHorizon")

    @coupon_paid_at_horizon.setter
    def coupon_paid_at_horizon(self, value):
        self._set_parameter("couponPaidAtHorizon", value)

    @property
    def haircut_rate_percent(self):
        """
        Repo contract's haircut par in percentage, for computing the transaction's Margin. InitialMargin = 1/ 1-Haircut.
        Optional. Either HairCut or Initial Marging field can be bet.
        :return: float
        """
        return self._get_parameter("haircutRatePercent")

    @haircut_rate_percent.setter
    def haircut_rate_percent(self, value):
        self._set_parameter("haircutRatePercent", value)

    @property
    def initial_margin_percent(self):
        """
        Repo contract's initial margin in percentage.
        :return: float
        """
        return self._get_parameter("initialMarginPercent")

    @initial_margin_percent.setter
    def initial_margin_percent(self, value):
        self._set_parameter("initialMarginPercent", value)



# coding: utf8

from typing import Optional, Iterable, List

from ._repo_definition import RepoInstrumentDefinition
from ._repo_pricing_parameters import PricingParameters
from ._repo_underlying_contract import UnderlyingContract
from .._base_definition import BaseDefinition
from ..._enums import DayCountBasis
from ....._tools import create_repr


class Definition(BaseDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    instrument_tag : str, optional
        User defined string to identify the instrument.It can be used to link output
        results to the instrument definition. Only alphabetic, numeric and '- _.#=@'
        characters are supported.
    start_date : str, optional
        Start date of the repo, that means when the the underlying security is
        exchanged.
    end_date : str, optional
        End date of the repo, that means when the borrower repurchases the security
        back. Either end_date or tenor field are requested.
    tenor : str, optional
        tenor that defines the duration of the Repo in case no end_date has been
        provided. In that case, end_date is computed from start_date and tenor. Either
        end_date or tenor field are requested.
    day_count_basis : DayCountBasis, optional
        Day Count Basis convention to apply to the custom Repo rate.
        By default "Dcb_Actual_360".
    underlying_instruments : list of UnderlyingContract
        Definition of the underlying instruments. Only Bond Contracts are supported for
        now, and only one Bond can be used.
    is_coupon_exchanged : bool, optional
        Specifies whether or not intermediate coupons are exchanged.
        - CouponExchanged = True to specify that intermediate coupons for the underlying
          bond (between the repo start date and repo end date) are exchanged between the
          repo seller and repo buyer.
        - CouponExchanged = False to specify that no intermediate coupons are exchanged
          between the repo seller and repo buyer. In this case the repo instrument is
          like a standard loan with no intermediate coupons; the bond is only used as a
          warranty in case the money borrower defaults. True by default, which means
          coupon exchanged.
    repo_rate_percent : float, optional
        Custom Repo Rate in percentage. If not provided in the request, it will be
        computed by interpolating/extrapolating a Repo Curve.
    fields: list of str, optional
        Contains the list of Analytics that the quantitative analytic service will
        compute.
    pricing_parameters : PricingParameters, optional
        The pricing parameters to apply to this instrument. If pricing parameters
        are not provided at this level parameters defined globally at the request
        level are used. If no pricing parameters are provided globally default
        values apply.
    extended_params : dict, optional
        If necessary other parameters

    Methods
    -------
    get_data(session=session, on_response=on_response, async_mode=None)
        Returns a response to the data platform
    get_data_async(session=session, on_response=on_response, async_mode=None)
        Returns a response to the async data platform
    get_stream(session=session, api="")
        Get stream quantitative analytic service subscription

    Examples
    --------
     >>> import refinitiv.data.content.ipa.financial_contracts as rdf
     >>> definition = rdf.repo.Definition(
     ...   underlying_instruments=[
     ...   rdf.repo.UnderlyingContract(
     ...   instrument_type="Bond",
     ...   instrument_definition=rdf.bond.Definition(instrument_code="US191450264="),
     ...   )],
     ...)
     >>> response = definition.get_data()

     Using get_stream
     >>> stream = definition.get_stream()
     >>> stream.open()
    """

    def __init__(
        self,
        *,
        instrument_tag: Optional[str] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        tenor: Optional[str] = None,
        day_count_basis: Optional[DayCountBasis] = None,
        underlying_instruments: Optional[List[UnderlyingContract]] = None,
        is_coupon_exchanged: Optional[bool] = None,
        repo_rate_percent: Optional[float] = None,
        fields: Optional[Iterable[str]] = None,
        pricing_parameters: Optional[PricingParameters] = None,
        extended_params: Optional[dict] = None,
    ) -> None:
        self.underlying_instruments = underlying_instruments
        definition = RepoInstrumentDefinition(
            instrument_tag=instrument_tag,
            underlying_instruments=underlying_instruments,
            start_date=start_date,
            end_date=end_date,
            tenor=tenor,
            is_coupon_exchanged=is_coupon_exchanged,
            repo_rate_percent=repo_rate_percent,
            day_count_basis=day_count_basis,
        )
        super().__init__(
            definition=definition,
            fields=fields,
            pricing_parameters=pricing_parameters,
            extended_params=extended_params,
        )

    def __repr__(self):
        return create_repr(
            self,
            middle_path="content.ipa.financial_contract",
            content=f"{{underlying_instruments='{self.underlying_instruments}'}}",
        )



from typing import Iterable, Union

from .._content_provider import ContentProviderLayer
from .._content_type import ContentType
from ..._tools import create_repr, validate_bool_value


class Definition(ContentProviderLayer):
    """
    This class describe parameters to retrieve ESG basic data.

    Parameters
    ----------
    universe : str, list of str
        The Universe parameter allows the user to define the company they
        want content returned for, ESG content is delivered at the Company Level.

    closure : str, optional
        Specifies the parameter that will be merged with the request

    use_field_names_in_headers: bool, optional
        Return field name as column headers for data instead of title

    Examples
    --------
    >>> from refinitiv.data.content import esg
    >>> definition = esg.basic_overview.Definition("IBM.N")
    >>> response = definition.get_data()

    >>> response = await definition.get_data_async()
    """

    def __init__(
        self,
        universe: Union[str, Iterable[str]],
        closure: str = None,
        use_field_names_in_headers: bool = False,
    ):
        validate_bool_value(use_field_names_in_headers)

        super().__init__(
            content_type=ContentType.ESG_BASIC_OVERVIEW,
            universe=universe,
            closure=closure,
            use_field_names_in_headers=use_field_names_in_headers,
        )

    def __repr__(self):
        return create_repr(
            self,
            middle_path="basic_overview",
            content=f"{{universe='{self._kwargs.get('universe')}'}}",
        )



# coding: utf-8
import requests

from .event import StreamEvent
from .stream_connection import StreamConnection, LOGIN_STREAM_ID
from .stream_cxn_state import StreamCxnState


class RDPStreamConnection(StreamConnection):
    @property
    def subprotocol(self) -> str:
        return "rdp_streaming"

    def get_login_message(self):
        return self.session.get_rdp_login_message(LOGIN_STREAM_ID)

    def _handle_login_message(self, message: dict):
        """
        Parameters
        ----------
        message
            {
                'state': {
                    'code': 200,
                    'status': 'OK',
                    'message': 'Access token is valid'
                },
                'type': 'Ack',
                'streamID': '2'
            }
        """
        state = message.get("state", {})
        status = state.get("status")
        code = state.get("code")

        # "OK" for qps and "Ok" for tds
        if status == "OK" or status == "Ok":
            self._state = StreamCxnState.Ready
            self.prepared.set_result(True)

        elif status == "Closed":
            self.error(
                f"{self._classname} received a closing message "
                f"(cxn state={self.state}, message state={state})"
            )
            self._state = StreamCxnState.Disconnected
            self.dispose()

        elif status == "Error" or code == requests.codes.bad:
            self.error(
                f"{self._classname} received an error message "
                f"(cxn state={self.state}, message state={state})"
            )
            self._state = StreamCxnState.Disconnected
            self.dispose()

        else:
            raise ValueError(
                f"{self._classname} Don't know what to do "
                f"(cxn state={self.state}, "
                f"message status={status}, message={message})"
            )

    def _process_message(self, message: dict) -> None:
        self.debug(f"{self._classname} process_message {message} ")

        stream_id = message.get("streamID")
        message_type = message.get("type")

        event = StreamEvent.get(stream_id)

        if message_type == "Ack":
            self._emitter.emit(event.ack_by_id, self, message)

        elif message_type == "Response":
            self._emitter.emit(event.response_by_id, self, message)

        elif message_type == "Update":
            self._emitter.emit(event.update_by_id, self, message)

        elif message_type == "Alarm":
            self._emitter.emit(event.alarm_by_id, self, message)

        elif message_type == "Error":
            self._emitter.emit(event.error_by_id, self, message)

        elif message_type == "Heartbeat":
            # do nothing
            pass

        else:
            raise ValueError(f"Unknown message {message}")



# coding: utf8


def _convert_to_str(from_type, values, some):
    result = None

    if isinstance(some, str):
        result = from_type.normalize(some)
    elif isinstance(some, from_type):
        result = some.value

    if result:
        return result
    else:
        raise TypeError(f"{from_type.__name__} value must be in {values}")


def _normalize(lower_values, some):
    some_lower = some.lower()
    enum_some = lower_values.get(some_lower)
    result = ""

    if enum_some:
        result = enum_some.value
    return result


def is_enum_equal(enum_value, some):
    c = enum_value.__class__
    try:
        s = c.convert_to_str(some)
        is_equal = enum_value.value == s
    except TypeError:
        is_equal = False
    return is_equal



import math
import threading
from typing import Callable, Optional, Union
from typing import TYPE_CHECKING

import pandas as pd

from ._ohlc_builder import (
    merge_dataframes,
    create_df,
    retrieve_data_for_df,
    Ticks_OHLCBuilder,
    OHLCBuilder,
)
from ._recording_control import (
    RecordingControl,
    NoBlocking_RecordingControl,
    Blocking_RecordingControl,
    RepeatNonBlocking_RecordingControl,
    RepeatBlocking_RecordingControl,
)
from ._stream_update_handler import (
    BuildDF_StreamUpdateHandler,
    CollectUpdates_StreamUpdateHandler,
    StreamUpdateHandler,
)
from ..session import get_default

if TYPE_CHECKING:
    from ..content._streamingprices import StreamingPrices


class PricingRecorder:
    """
    Pricing recorder class allows to record updates from server.
    Create dataframes based on received updates.

    Parameters
    ----------

    stream : StreamingPrices
        StreamingPrices object

    """

    def __init__(self, stream: "StreamingPrices"):
        self._stream = stream
        self._logger = get_default().logger()
        self._event = threading.Event()

        self._frequency = None
        self._duration = None
        self._ticks_per_bar = None
        self._on_data = None
        self._record_called = 0

        self.is_running = False

        self._ohlc_builder: Union[OHLCBuilder, None] = None
        self._update_handler: Optional[StreamUpdateHandler] = None
        self._recording_control: Optional[RecordingControl] = None

    @staticmethod
    def _parse_input_frequency_and_duration(value: str) -> int:
        last_char = value[-1]
        if last_char not in ["s", "h", "d", "m"] and not value.endswith("min"):
            raise ValueError(
                "Please provide 'duration' or 'frequency' value in "
                "valid format. For example: '10s', '2min', '1h'"
            )

        result = None
        try:
            if "s" == last_char:
                result = int(value[:-1])
            elif value.endswith("min"):
                seconds = int(value[:-3])
                result = seconds * 60
            elif "h" == last_char:
                seconds = int(value[:-1])
                result = seconds * 3600
            elif "d" == last_char:
                seconds = int(value[:-1])
                result = seconds * 3600 * 24
            elif "m" == last_char:
                seconds = int(value[:-1])
                result = seconds * 3600 * 24 * 30
        except ValueError:
            raise ValueError(
                "Please provide 'duration' value"
                " in valid format. For example: '10s', '2min', '1h'"
            )

        return result

    @staticmethod
    def _validate_count_argument(ticks_per_bar):
        try:
            ticks_per_bar = int(ticks_per_bar)
        except ValueError:
            raise ValueError(
                "Invalid argument. Please provide 'ticks_per_bar'"
                " in the following format: '10', '100', '500'"
            )

        if ticks_per_bar <= 0:
            raise ValueError("Invalid argument. 'ticks_per_bar' should be more then 0")

    def _validate_arguments(self, frequency: str, duration: str, ticks_per_bar: str):
        if ticks_per_bar != "1" and frequency != "tick":
            raise ValueError(
                "Please provide 'tick' value as frequency when you are using 'ticks_per_bar' argument."
            )

        self._validate_count_argument(ticks_per_bar)

        if duration and frequency != "tick":
            frequency = self._parse_input_frequency_and_duration(frequency)
            duration = self._parse_input_frequency_and_duration(duration)

            self._expected_count_of_callbacks = duration / frequency

            float_part, self._expected_count_of_callbacks = math.modf(
                self._expected_count_of_callbacks
            )

            self._callback_called_count = 0
            if duration % frequency:
                self._expected_count_of_callbacks += 1

            if frequency > duration:
                raise ValueError(
                    "Please check your arguments, 'duration'"
                    " should be higher that 'frequency'."
                )

    @staticmethod
    def _check_df(df: pd.DataFrame) -> pd.DataFrame:
        if isinstance(df, pd.DataFrame):
            df.fillna(pd.NA, inplace=True)

        else:
            df = pd.DataFrame()

        return df

    def record(
        self,
        frequency: str = "tick",
        duration: "str" = None,
        ticks_per_bar: str = "1",
        on_data: Callable = None,
    ):
        """
        Starts recording updates from server and save it in memory

        Parameters
        ----------
        frequency : str, optional
            Using to calculate ohlc based on received updates during period that was provided
        duration : str, optional
            Duration of recording data. Could be provided in seconds, minutes, hours
        ticks_per_bar : str, optional
            Count of ticks to record
        on_data : function, optional
             Callback which is calling with 'frequency' and receive dataframe
             with calculated ohlc from last updates and recorder object.
             Frequency has to be provided

        Returns
        -------
        Examples
        -------
        Start recording all updates during 15 seconds and calculate ohlc

        >>> import refinitiv.data as rd
        >>> stream = rd.open_pricing_stream(universe=['EUR='], fields=['BID', 'ASK', 'OPEN_PRC'])
        >>> stream.recorder.record(duration="15s")
        >>> stream.recorder.stop()

        >>> stream.close()
        >>> history = stream.recorder.get_history()
        >>> history.ohlc("5s")

        Start recording updates and calculate ohlc
        by using 'frequency' and call 'callback'
        function with updated ohlc dataframe every 5 seconds
        >>> import refinitiv.data as rd
        >>> stream = rd.open_pricing_stream(universe=['EUR='], fields=['BID', 'ASK', 'OPEN_PRC'])
        >>>
        >>>
        >>> def callback(dataframe, recorder):
        ...     print(dataframe)
        >>>
        >>> stream.recorder.record(frequency="5s", duration="15s", on_data=callback)
        >>> stream.recorder.stop()
        >>> stream.close()
        >>> history = stream.recorder.get_history()

        """
        if self._stream.is_closed:
            raise ConnectionError("Stream is closed. Cannot record.")

        self._validate_arguments(frequency, duration, ticks_per_bar)

        self._frequency = frequency
        self._duration = duration
        self._ticks_per_bar = int(ticks_per_bar)
        self._on_data = on_data

        self.is_running = True

        frequency_tick = self._frequency == "tick"
        no_duration = not self._duration
        ticks_1 = self._ticks_per_bar == 1
        ticks_not_1 = self._ticks_per_bar != 1

        # stream.recorder.record(frequency='tick')
        if all([frequency_tick, no_duration, ticks_1]):
            if not isinstance(self._update_handler, CollectUpdates_StreamUpdateHandler):
                self._update_handler = CollectUpdates_StreamUpdateHandler(
                    self._stream, recorder=self
                )
            self._recording_control = NoBlocking_RecordingControl(self._update_handler)
            self._recording_control.start_recording()

        # stream.recorder.record(frequency='tick', ticks_per_bar=10)
        elif all([frequency_tick, no_duration, ticks_not_1]):
            self._create_ohlc_builder(Ticks_OHLCBuilder)
            self._update_handler = BuildDF_StreamUpdateHandler(
                self._stream,
                self._ohlc_builder,
                self._ticks_per_bar,
                recorder=self,
                on_record_callback=on_data,
            )
            self._recording_control = NoBlocking_RecordingControl(self._update_handler)
            self._recording_control.start_recording()

        # stream.recorder.record(frequency='tick', duration="60s")
        elif all([frequency_tick, duration, ticks_1]):

            if not isinstance(self._update_handler, CollectUpdates_StreamUpdateHandler):
                self._update_handler = CollectUpdates_StreamUpdateHandler(
                    self._stream, recorder=self
                )

            self._recording_control = Blocking_RecordingControl(self._update_handler)
            interval = self._parse_input_frequency_and_duration(self._duration)
            self._recording_control.start_recording(interval)
            self.stop()

        # stream.recorder.record(frequency='tick', duration="60s", ticks_per_bar=10)
        elif all([frequency_tick, duration, ticks_not_1]):
            self._create_ohlc_builder(Ticks_OHLCBuilder)
            self._update_handler = BuildDF_StreamUpdateHandler(
                self._stream,
                self._ohlc_builder,
                self._ticks_per_bar,
                self,
                self._on_data,
            )
            self._recording_control = Blocking_RecordingControl(self._update_handler)
            interval = self._parse_input_frequency_and_duration(self._duration)
            self._recording_control.start_recording(interval)
            self.stop()

        # stream.recorder.record(frequency='5s')
        elif all([frequency, no_duration, ticks_1]):
            self._create_ohlc_builder(OHLCBuilder)
            self._update_handler = CollectUpdates_StreamUpdateHandler(
                self._stream, recorder=self
            )
            self._recording_control = RepeatNonBlocking_RecordingControl(
                self._on_data,
                self._update_handler,
                self._ohlc_builder,
                self._logger,
                self,
            )
            interval = self._parse_input_frequency_and_duration(self._frequency)
            self._recording_control.start_recording(interval)

        # stream.recorder.record(frequency='5s', duration="17s")
        elif all([frequency, duration, ticks_1]):
            self._create_ohlc_builder(OHLCBuilder)
            self._update_handler = CollectUpdates_StreamUpdateHandler(
                self._stream, recorder=self
            )
            duration = self._parse_input_frequency_and_duration(self._duration)
            self._recording_control = RepeatBlocking_RecordingControl(
                duration,
                self._on_data,
                self._update_handler,
                self._ohlc_builder,
                self._logger,
                self,
            )
            interval = self._parse_input_frequency_and_duration(self._frequency)
            self._recording_control.start_recording(interval)
            self.stop()

        else:
            raise ValueError(
                f"Cannot cover case when "
                f"frequency={self._frequency}, "
                f"duration={self._duration}, "
                f"ticks_per_bar={self._ticks_per_bar}"
            )

    def _create_ohlc_builder(self, klass):
        if not isinstance(self._ohlc_builder, klass):
            self._ohlc_builder = klass(
                self._frequency, self._stream.universe, self._stream.fields
            )

    def get_history(self) -> "pd.DataFrame":
        dfs = []

        if self._frequency == "tick" and self._ticks_per_bar == 1:

            updates_by_stream_name = self._update_handler.updates_by_stream_name
            for universe, stream_data in updates_by_stream_name.items():
                timestamps, data, fields = retrieve_data_for_df(stream_data)
                dataframe = create_df(data, timestamps, fields, universe)
                dfs.append(dataframe)

            if not dfs:
                msg = "We didn't receive any updates. Dataframe couldn't be created."
                self._logger.warning(msg)
                df = pd.DataFrame()

            else:
                df = merge_dataframes(dfs)

        else:
            df = self._check_df(self._ohlc_builder.ohlc_df)

        return df

    def stop(self):
        """
        Stop recording updates and cancel repeat timer for creating ohlc dataframes.
        """
        if not self.is_running:
            return

        self.is_running = False
        self._recording_control.stop_recording()

    def delete(self):
        """Delete whole recorded updates"""
        self._recording_control.delete_recording()



from ..._ipa_content_provider import IPAContentProviderLayer
from ..._surfaces._fx_surface_request_item import FxSurfaceRequestItem
from ..._surfaces._surface_types import SurfaceParameters, SurfaceLayout
from ...._content_type import ContentType
from ...._types import ExtendedParams, OptStr
from ....._tools import create_repr


class Definition(IPAContentProviderLayer):
    """
    Create a Fx data Definition object.

    Parameters
    ----------
    surface_layout : SurfaceLayout
        See details in SurfaceLayout class
    surface_parameters : SurfaceParameters
        See details in SurfaceParameters class
    underlying_definition : dict
       Dict with instrument_code
       Example:
            {"fxCrossCode": "EURUSD"}
    surface_tag : str, optional
        A user defined string to describe the volatility surface
    extended_params : dict, optional
        If necessary other parameters

    Methods
    -------
    get_data(session=session, on_response=on_response, **kwargs)
        Returns a response to the data platform
    get_data_async(session=None, on_response=None, **kwargs)
        Returns a response asynchronously to the data platform

    Examples
    --------
    >>> from refinitiv.data.content.ipa.surfaces import surface_fx
    >>> definition = surface_fx.Definition(
    ...    underlying_definition={"fxCrossCode": "EURUSD"},
    ...    surface_tag="FxVol-EURUSD",
    ...    surface_layout=surface_fx.SurfaceLayout(
    ...    format=surface_fx.Format.MATRIX
    ...    ),
    ...    surface_parameters=surface_fx.FxCalculationParams(
    ...    x_axis=surface_fx.Axis.DATE,
    ...    y_axis=surface_fx.Axis.STRIKE,
    ...    calculation_date="2018-08-20T00:00:00Z"
    ...    )
    ...)
    >>>
    """

    def __init__(
        self,
        surface_layout: SurfaceLayout,
        surface_parameters: SurfaceParameters,
        underlying_definition: dict,
        surface_tag: OptStr = None,
        extended_params: ExtendedParams = None,
    ):
        request_item = FxSurfaceRequestItem(
            surface_layout=surface_layout,
            surface_parameters=surface_parameters,
            underlying_definition=underlying_definition,
            surface_tag=surface_tag,
        )
        super().__init__(
            content_type=ContentType.SURFACES,
            universe=request_item,
            extended_params=extended_params,
        )

    def __repr__(self):
        return create_repr(self)



# coding: utf8

from typing import Optional

from .._instrument_definition import InstrumentDefinition
from ._enums import (
    BuySell,
    CallPut,
    ExerciseStyle,
    UnderlyingType,
)


class OptionDefinition(InstrumentDefinition):
    """
    Parameters
    ----------
    instrument_tag : str, optional
        User defined string to identify the instrument.It can be used to link output
        results to the instrument definition. Only alphabetic, numeric and '- _.#=@'
        characters are supported. Optional.
    end_date : str, optional
        Expiry date of the option
    buy_sell : BuySell, optional
        The side of the deal. Possible values:
        - Buy
        - Sell
    call_put : CallPut, optional
        Tells if the option is a call or a put. Possible values:
        - Call
        - Put
    exercise_style : ExerciseStyle, optional
        EURO or AMER
    underlying_type : UnderlyingType, optional
        Underlying type of the option. Possible values:
        - Eti
        - Fx
    strike : float, optional
        strike of the option
    """

    def __init__(
        self,
        instrument_tag: Optional[str] = None,
        end_date: Optional[str] = None,
        buy_sell: Optional[BuySell] = None,
        call_put: Optional[CallPut] = None,
        exercise_style: Optional[ExerciseStyle] = None,
        underlying_type: Optional[UnderlyingType] = None,
        strike: Optional[float] = None,
        **kwargs,
    ) -> None:
        super().__init__(instrument_tag, **kwargs)
        self.end_date = end_date
        self.buy_sell = buy_sell
        self.call_put = call_put
        self.exercise_style = exercise_style
        self.underlying_type = underlying_type
        self.strike = strike

    @classmethod
    def get_instrument_type(cls):
        return "Option"

    @property
    def buy_sell(self):
        """
        The side of the deal. Possible values:
        - Buy
        - Sell
        :return: enum BuySell
        """
        return self._get_enum_parameter(BuySell, "buySell")

    @buy_sell.setter
    def buy_sell(self, value):
        self._set_enum_parameter(BuySell, "buySell", value)

    @property
    def call_put(self):
        """
        Tells if the option is a call or a put. Possible values:
        - Call
        - Put
        :return: enum CallPut
        """
        return self._get_enum_parameter(CallPut, "callPut")

    @call_put.setter
    def call_put(self, value):
        self._set_enum_parameter(CallPut, "callPut", value)

    @property
    def exercise_style(self):
        """
        EURO or AMER
        :return: enum ExerciseStyle
        """
        return self._get_enum_parameter(ExerciseStyle, "exerciseStyle")

    @exercise_style.setter
    def exercise_style(self, value):
        self._set_enum_parameter(ExerciseStyle, "exerciseStyle", value)

    @property
    def underlying_type(self):
        """
        Underlying type of the option. Possible values:
        - Eti
        - Fx
        :return: enum UnderlyingType
        """
        return self._get_enum_parameter(UnderlyingType, "underlyingType")

    @underlying_type.setter
    def underlying_type(self, value):
        self._set_enum_parameter(UnderlyingType, "underlyingType", value)

    @property
    def end_date(self):
        """
        Expiry date of the option
        :return: str
        """
        return self._get_parameter("endDate")

    @end_date.setter
    def end_date(self, value):
        self._set_parameter("endDate", value)

    @property
    def strike(self):
        """
        strike of the option
        :return: float
        """
        return self._get_parameter("strike")

    @strike.setter
    def strike(self, value):
        self._set_parameter("strike", value)



# coding: utf8

__all__ = ["BarrierType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class BarrierType(Enum):
    KNOCK_IN = "KnockIn"
    KNOCK_IN_KNOCK_OUT = "KnockInKnockOut"
    KNOCK_OUT = "KnockOut"
    KNOCK_OUT_KNOCK_IN = "KnockOutKnockIn"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(BarrierType, _BARRIER_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_BARRIER_TYPE_VALUES_IN_LOWER_BY_BARRIER_TYPE, some)


_BARRIER_TYPE_VALUES = tuple(t.value for t in BarrierType)
_BARRIER_TYPE_VALUES_IN_LOWER_BY_BARRIER_TYPE = {
    name.lower(): item for name, item in BarrierType.__members__.items()
}



# coding: utf8

__all__ = ["Direction"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class Direction(Enum):
    """
    The direction of the leg.

    The possible values are:
        Paid - the cash flows of the leg are paid to the counterparty.
        Received - the cash flows of the leg are received from the counterparty.
    """

    PAID = "Paid"
    RECEIVED = "Received"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(Direction, _DIRECTION_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_DIRECTION_VALUES_IN_LOWER_BY_DIRECTION, some)


_DIRECTION_VALUES = tuple(t.value for t in Direction)
_DIRECTION_VALUES_IN_LOWER_BY_DIRECTION = {
    name.lower(): item for name, item in Direction.__members__.items()
}



import pandas as pd

from ..._tools._dataframe import (
    convert_column_df_to_datetime,
    convert_items_df_to_datetime,
)
from ...eikon._tools import tz_replacer


def news_build_df_udf(raw, **kwargs):
    df = convert_content_data_to_df_udf(raw)
    return df


def news_build_df_rdp(raw, **kwargs):
    columns = ["versionCreated", "text", "storyId", "sourceCode"]

    if isinstance(raw, list):
        data = []

        for i in raw:
            data.extend(i["data"])

    else:
        data = raw["data"]

    first_created = [
        tz_replacer(headline["newsItem"]["itemMeta"]["firstCreated"]["$"])
        for headline in data
    ]
    index = convert_items_df_to_datetime(first_created)
    headlines = []

    for headline_data in data:
        news_item = headline_data.get("newsItem", dict())
        item_meta = news_item.get("itemMeta", {})
        info_sources = news_item["contentMeta"]["infoSource"]
        info_source = next(
            (
                item["_qcode"]
                for item in info_sources
                if item["_role"] == "sRole:source"
            ),
            None,
        )
        version_created = pd.to_datetime(item_meta["versionCreated"]["$"])
        headlines.append(
            [
                version_created,
                item_meta["title"][0]["$"],
                headline_data["storyId"],
                info_source,
            ]
        )
    if headlines:
        dataframe = pd.DataFrame(
            data=headlines,
            index=index,
            columns=columns,
        )

        if not dataframe.empty:
            dataframe = dataframe.convert_dtypes()
        dataframe["versionCreated"] = dataframe["versionCreated"].dt.tz_localize(None)

    else:
        dataframe = pd.DataFrame([], columns=columns)
    return dataframe


def convert_content_data_to_df_udf(content_data: dict) -> pd.DataFrame:
    selected_fields = ["versionCreated", "text", "storyId", "sourceCode"]

    raw_headlines = content_data.get("headlines", [])
    first_created = [
        tz_replacer(headline["firstCreated"]) for headline in raw_headlines
    ]
    index = convert_items_df_to_datetime(first_created)

    headlines = [
        [headline[field] for field in selected_fields] for headline in raw_headlines
    ]
    if len(headlines):
        df = pd.DataFrame(
            headlines,
            index=index,
            columns=selected_fields,
        )

        if not df.empty:
            df = df.convert_dtypes()

    else:
        df = pd.DataFrame([], index, selected_fields)

    convert_column_df_to_datetime(df, "versionCreated")
    df.fillna(pd.NA, inplace=True)

    return df


def _get_text_from_story(story):
    news_item = story.get("newsItem", dict())
    content_set = news_item.get("contentSet", dict())
    inline_data = content_set.get("inlineData", [dict()])
    return inline_data[0].get("$")


def _get_headline_from_story(story):
    news_item = story.get("newsItem", dict())
    content_meta = news_item.get("contentMeta", dict())
    headline = content_meta.get("headline", [dict()])
    return headline[0].get("$")


def get_headlines_rdp(raw, create_headline_func, limit):
    headlines = []

    if isinstance(raw, list):

        data = []
        for i in raw:
            data.extend(i.get("data", i.get("headlines", [])))

    else:
        data = raw.get("data", raw.get("headlines", []))

    for datum in data:
        headline = create_headline_func(datum)
        headlines.append(headline)

    headlines = headlines[:limit]
    return headlines
